# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Convenient timers
=================

Usage::
  import fife.extensions.fife_timer

  fife_timer.init( my_fife_engine.getTimeManager() )

  def spam():
     print "SPAM SPAM ",

  repeater = fife_timer.repeatCall(500,spam)

  def stop_spam():
     repeater.stop()
     print "BACON EGGS AND SPAM"

  delayed = fife_timer.delayCall(50000,stop_spam)

"""

from fife import fife

#global time manager
_manager = None

def init(timemanager):
	"""
	Initialize timers.

	@param timemanager: A L{fife.TimeManager} as retuned by L{fife.Engine.getTimeManager}.
	"""
	global _manager
	_manager = timemanager

class Timer(fife.TimeEvent):
	"""
	Timer

	This class wraps the fife.TimeEvent class to make it easily usable from Python
	It allows for a TimeEvent to be executed once or multiple times.

	Remember FIFE::TimeManager does NOT delete the timer so make sure you keep a reference
	to this timer to ensure python doesnt delete the timer prematurely.
	"""
	def __init__(self,delay=0,callback=None,repeat=0):
		"""
		@param delay: The delay in milliseconds to execute the callback
		@param callback: The function to execute when the time delay has passed
		@param repeat: The number of times to execute the callback.  1=once, 0=forever
		"""
		super(Timer,self).__init__(delay)
		self._active = False
		self._callback = callback
		self._manager = _manager
		self.setPeriod(delay)
		self._repeat = repeat
		self._executed = 0

	def start(self):
		"""
		Call this to start the timer.

		This registers the timer with the time manager.  The time manger then
		calls the timers updateEvent() function when the delay time has passed.
		"""
		if self._active:
			return
		self._active = True

		self._executed = 0

		self.setLastUpdateTime(self._manager.getTime())
		self._manager.registerEvent(self)

	def stop(self):
		"""
		Stops the timer

		This unregisters the timer from the time manager.
		"""
		if not self._active:
			return

		self._active = False
		self._manager.unregisterEvent(self)

	def updateEvent(self,delta):
		"""
		This is called by FIFE::TimeManager when the delay has passed.

		Should not be called directly.
		"""

		if self._repeat != 0:
			self._executed += 1
			if self._executed >= self._repeat:
				self.stop()

		if callable(self._callback):
			self._callback()

	def _setDelay(self, delay):
		"""
		Sets how many milliseconds to wait before executing the callback.

		The timer must not be active to change this value

		@param delay: Number of milliseconds to wait before executing the callback.
		@type delay: C{integer}
		"""

		if not self._active:
			self.setPeriod(delay)

	def _getDelay(self):
		"""
		Returns the number of milliseconds to wait before executing the callback.

		@return: Number of milliseconds.
		@rtype: C{integer}
		"""
		return self.getPeriod()

	def _setCallback(self, callback):
		self._callback = callback

	def _getCallback(self):
		return self._callback

	def _setRepeat(self, repeat):
		"""
		Sets how many times the timer should be repeated.

		The timer must not be active to change it's repeat value.

		@param repeat: Number of times to repeat the timer.  0=forever, 1=once.
		@type repeat: C{integer}
		"""

		if not self._active:
			self._repeat = repeat

	def _getRepeat(self, repeat):
		"""
		Returns the number of times the timer will be executed.

		@return: Number of times the timer will be executed.
		@rtype: C{integer}
		"""
		return self._repeat

	def _getActive(self):
		"""
		Returns True if the timer is active and False if it is not.

		@return: True if timer is active, False if it is not.
		@rtype: C{boolean}
		"""
		return self._active

	def _getNumExecuted(self):
		"""
		Returns the number of times the timer has been executed

		@return: Number of times the timer has been executed
		@rtype: C{integer}
		"""
		return self._executed

	delay = property(_getDelay, _setDelay)
	callback = property(_getCallback, _setCallback)
	repeat = property(_getRepeat, _setRepeat)
	active = property(_getActive)
	numexecuted = property(_getNumExecuted)


def delayCall(delay,callback):
	"""
	Delay a function call by a number of milliseconds.

	Remember to keep a reference to the timer this function returns.  If you
	do not python will delete the timer prematurely which may case a segfault.

	@param delay: Delay in milliseconds.
	@param callback: The function to call.

	@return: The timer.
	@rtype: L{Timer}
	"""
	timer = Timer(delay, callback, 1)
	timer.start()
	return timer


def repeatCall(period,callback):
	"""
	Repeat a function call. The call is repeated until the timer is stopped.

	Remember to keep a reference to the timer this function returns.  If you
	do not python will delete the timer prematurely which may case a segfault.

	@param period: Period between calls in milliseconds.
	@param callback: The function to call.

	@return: The timer.
	@rtype: L{Timer}
	"""
	timer = Timer(period, callback, 0)
	timer.start()
	return timer

__all__ = ['init','Timer','delayCall','repeatCall']




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

class PyChanException(Exception):
	"""
	Base exception class for PyChan.
	All exceptions raised by PyChan derive from this.
	"""
	pass

class InitializationError(PyChanException):
	"""
	Exception raised during the initialization.
	"""
	pass

class RuntimeError(PyChanException):
	"""
	Exception raised during the run time - for example caused by a missing name attribute in a XML file.
	"""
	pass

class GuiXMLError(PyChanException):
	"""
	An error that occured during parsing an XML file.
	"""

class ParserError(PyChanException):
	"""
	An error that occured during parsing an attribute.
	"""

class PrivateFunctionalityError(RuntimeError):
	"""
	Exception raised if private attributes/functions are used.
	"""

class StopTreeWalking(StopIteration):
	"""
	Internal exception used to abort iteration over the widget tree.
	"""

class DeprecatedException(PyChanException):
	"""
	Exception raised if a deprecated function or format is used.
	"""



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
FIFE Backwards Combatibility Layer
==================================

This module can be imported if you want to
run code that wasn't adapted to API changes in FIFE.

2008.1
------

 - Animation.addFrame now expects a fife.ResourcePtr instead of an fife.Image
 - Pool.getIndex is just an alias for Pool.addResourceFromFile.
 - EventManager.setNonConsumableKeys is superseeded by EventManager.setKeyFilter

"""
from __future__ import print_function

from fife import fife

# Utility functions

def deprecated(revision,message):
	print("fife_compat: Deprecation warning - See revision %d " % revision)
	print(" - ",message)

def this_is_deprecated(func,revision=0,message=None):
	if message is None:
		message = repr(func) + " is deprecated."
	def wrapped_func(*args,**kwargs):
		deprecated(revision,message)
		return func(*args,**kwargs)
	return wrapped_func

def _compat_NonConsumableKeys():
	class CompatKeyFilter(fife.IKeyFilter):
		def __init__(self, keys):
			fife.IKeyFilter.__init__(self)
			self.keys = keys

		def isFiltered(self, event):
			return event.getKey().getValue() in self.keys

	def _setNonConsumableKeys(self,keys):
		deprecated(2636, "Write an IKeyFilter instead of using EventManager.setNonConsumableKeys.\n" +
				 "You probably don't need it anyway")
		self.compat_keyfilter = CompatKeyFilter(keys)
		self.compat_keyfilter.__disown__()
		self.setKeyFilter(self.compat_keyfilter)

	def _getNonConsumableKeys(self,keys):
		deprecated(2636, "Write an IKeyFilter instead of using EventManager.getNonConsumableKeys.")
		return self.compat_keyfilter.keys

	fife.EventManager.setNonConsumableKeys = _setNonConsumableKeys
	fife.EventManager.getNonConsumableKeys = _getNonConsumableKeys

_compat_NonConsumableKeys()



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""\
Pythonic Fifechan Wrapper - PyChan
=================================

Pythonic GUI API.

Features
--------
 - Simpler Interface
 - Very Basic XML Format support
 - Basic Layout Engine
 - Pseudo-Synchronous Dialogs.
 - Automagic background tiling (WIP)
 - Basic Styling support.
 - Simple Font Handling

TODO
----
 - Make setting parent attribute imply containment relation.
 - Finalize Widget.execute

 - Documentation ( Allways not enough :-( )
 - Handle Image Fonts
 - Move Font config files to XML, too ...

 - Implement real Menus
 - Implement StackWidget
 - Then implement TabWidget

 - GridLayout
 - Table

BUGS
----
 - Focus problems with Widget.execute.
 - Font.glyph_spacing is rendered incorrectly.
 - Is this a bug? At least inconvenient. MouseEntered events are not distributed for freshly shown widget.
 - It just looks bad.

Problems
--------
 - Reference counting problems again -sigh-
 ... and thus possible leaks.
 - High amount of code reuse -> Complex code
 - Needs at least new style classes and other goodies.
 - Missing documentation on:
     - Styling
     - ScrollArea
     - Fonts

How to use
==========

At its core you only need a few functions.
After setting up FIFE you need to initialize
pychan. After that you can load a GUI from an
XML file. Please see the documentation of L{loadXML}
for the details of the XML format
::
   import pychan
   pychan.init(fifeEngine)
   guiElement = pychan.loadXML("contents/gui/myform.xml")

The resulting guiElement can be shown and hidden with the
obvious L{widgets.Widget.show} and L{widgets.Widget.hide} methods.

To get a specific widget you have to give it a name in the XML
definition and use that to extract the widget from the returned
GUI element.
::
   okButton = guiElement.findChild(name="okButton")
   myInput = guiElement.findChild(name="myInput")

The data is extracted and set via direct attribute access.
These are using the python property technique to hide
behind the scenes manipulations. Please keep in mind that
the Layout engine and the exact way the widgets are displayed
is somewhat limited.
::
   myInput.text = "Blahblah"
   myList.items = ["1","2"]
   guiElement.position = (80,90)

A dialog without an OK button would be futile - so here's how
you hook widget events to function calls. Every widget
has a L{widgets.Widget.capture} method, which will directly call the passed
function when a widget event occurs. As a convenience a
L{widgets.Widget.mapEvents} function will batch the L{widgets.Widget.findChild} and
L{widgets.Widget.capture} calls in an obvious way.
::
   myButton.capture( application.quit )
   guiElement.mapEvents({
      'okButton' : self.applyAndClose,
      'closeButton':  guiElement.hide
   })

Other important places to look for information:
  - L{widgets.Widget} - Attributes explained.
  - L{loadXML} - Explain the XML format.
  - L{widgets.layout.LayoutBase} - Working of the layout engine.

Initialization, data distribution and collection
================================================

Very often a dialogs text fields, labels and listboxes have to be filled with data
after the creation of the dialog. This can be a tiresome process.
After a dialog has executed, B{other} attributes have to be read out again,
this to can be tiresome. PyChan simplifies both processes. But it treats them as three
processes. One is setting the data that will never be read out again - called B{initial data} -
the text of a checkbox or the list of a listBox are good examples. The second is setting the
data that is mutable by the user and may be read out again - for example the state of a checkbox
or the selected index in a list. The third and final process is collection of the user-mutable data::
  guiElement.distributeInitialData({
  	'myListBox' : choices,
  	'myLabel' : map.name,
  })
  guiElement.distributeData({
  	'myTextField' : map.description
  })
  # ... process dialog.
  map.description, choice = guiElement.collectData('myListBox','myTextField')
  print "You selected:",choice,", good choice!"

See L{widgets.Widget.distributeData},L{widgets.Widget.distributeInitialData},
L{widgets.Widget.collectData} and L{widgets.Widget.collectDataAsDict}.

Styling and font handling
=========================

Note: These features are B{work in progress} and likely to change.

A style is a set of rules for matching widgets and a set of attributes
applied to them after creation. The attributes can be any of the given
attributes. Matching happens currently only by the widget class name
itself.

As an example the following style - written as a python data structure -
will set the border size of all labels to 10::
  style = {
      'ListBox' : { # Matches all listboxes
           'border_size : 10 # After creation call  lisbox.border_size = 10
      }
  }

As a convenience you can use the string B{default} to match all widget
classes and thus - for example apply a common font::
  style = {
     'default' : {
          'font' : 'console_small'
     }
  }

A new style is added to pychan with L{internal.Manager.addStyle}.
You can set a new default style by adding a style with the name 'default'.

The font is set via a string identifier pulled from a font definition
in a PyChan configuration file. You have to load these by calling
L{loadFonts} in your startup code::
   import pychan
   pychan.init( fifeEngine )
   pychan.loadFonts( "content/fonts/console.xml" )

The font definition files are in XML format.
Unicode and internationalisation
================================

All text that is visible and editable by the player has to be a unicode object.
All text that is used internally, e.g. widget names, have to be normal strings.

While PyChan will not raise an exception, if you do not follow this guideline,
you are encouraged to so.

You can change the way unicode encoding errors are handled by using the
function L{setUnicodePolicy}.


Widget hierachy
===============

Every widget can be contained in another container widget like L{Window}, L{VBox},
L{HBox}, L{Container} or L{ScrollArea}. Container widgets can contain any number
of widgets. Thus we have a tree like structure of the widgets - which finally makes
up the window or frame that is placed on the screen.

In PyChan widgets are supposed to be manipulated via the root of this hierachy,
so that the actual layout can be changed in the XML files without hassle.
It can be compared to how HTML works.

These bits and pieces connect things up::
 -  name - A (hopefully) unique name in the widget hierachy
 -  findChildren - The accessor method to find widgets by name or any other attribute.
 -  _parent - The parent widget in the widget hierachy
 -  deepApply - The method used to walk over the widget hierachy. You have to reimplement
   this in case you want to provide custom widgets.

Wrapping machinery
==================

The wrapping mechanism works be redirecting attribute access to the Widget
derived classes to a C{real_widget} member variable which in turn is an instance
of the SWIG wrapped Fifechan widget.

To ensure the real widget has already been constructed, when the wrapping machinery
is already in use, this has to be the first attribute to set in the constructors.
This leads to a reversed construction sequence as the super classes constructor
has to be invoked I{after} the subclass specific construction has taken place.

"""
from __future__ import print_function
from __future__ import absolute_import
import sys

from builtins import map
__all__ = [
	'loadXML',
	'loadFonts',
	'init',
	'manager'
]

# This *import should really be removed!
from .widgets import *
from .widgets.tabbedarea import Tab
from .exceptions import *

from .fonts import loadFonts

### Initialisation ###

manager = None
def init(engine,debug=False, compat_layout=False):
	"""
	This has to be called before any other pychan methods can be used.
	It sets up a manager object which is available under pychan.manager.

	@param engine: The FIFE engine object.
	@param debug: bool - Enables and disables debugging output. Default is False.
	@param compat_layout: bool - Enables and disables compat layout. Default is False.
	"""
	from .compat import _munge_engine_hook
	from .internal import Manager
	global manager

	manager = Manager(_munge_engine_hook(engine),debug,compat_layout)

# XML Loader

from xml.sax import saxutils, handler
from traceback import print_exc

def traced(f):
	"""
	Simple decorator that prints tracebacks for any exceptions occuring in a
	function. Useful to avoid the infamous 'finally pops bad exception'
	that shadows the real cause of the error ...
	"""
	def traced_f(*args,**kwargs):
		try:
			return f(*args,**kwargs)
		except:
			print_exc()
			raise
	return traced_f

if sys.version_info < (3, ):
    class __GuiLoaderBase(object, handler.ContentHandler):
        pass
else:
    class __GuiLoaderBase(handler.ContentHandler):
        pass
    
class _GuiLoader(__GuiLoaderBase):
	def __init__(self):
		super(_GuiLoader,self).__init__()
		self.root = None
		self.indent = ""
		self.stack = []

	def _printTag(self,name,attrs):
		if not manager.debug: return
		attrstrings = ['%s="%s"' % tuple(map(str,t)) for t in list(attrs.items())]
		tag = "<%s " % name + " ".join(attrstrings) + ">"
		try:
			print(self.indent + tag)
		except UnicodeEncodeError as e:
			print(self.indent + tag.encode('ascii', 'backslashreplace'))

	def _resolveTag(self,name):
		""" Resolve a XML Tag to a PyChan GUI class. """
		cls = WIDGETS.get(name,None)
		if cls is None:
			raise GuiXMLError("Unknown GUI Element: %s" % name)
		return cls

	def _setAttr(self,obj,name,value):
		if not hasattr(obj.__class__,'ATTRIBUTES'):
			raise PyChanException("The registered widget/spacer class %s does not supply an 'ATTRIBUTES'."
								  % repr(obj))
		try:
			for attr in obj.ATTRIBUTES:
				if attr.name == name:
					attr.set(obj,value)
					return
		except GuiXMLError as e:
			raise GuiXMLError("Error parsing attr '%s'='%s' for '%s': '%s'" % (name,value,obj,e))
		raise GuiXMLError("Unknown GUI Attribute '%s' on '%s'" % (name,repr(obj)))

	def startElement(self, name, attrs):
		self._printTag(name,attrs)
		cls = self._resolveTag(name)
		if issubclass(cls,Widget):
			self.stack.append('gui_element')
			self._createInstance(cls,name,attrs)
		else:
			self.stack.append('unknown')
		self.indent += " "*4

	def _createInstance(self,cls,name,attrs):
		obj = cls(parent=self.root)
		for k,v in list(attrs.items()):
			self._setAttr(obj,k,v)

		if self.root:
			if isinstance(obj,Tab):
				if hasattr(self.root,'addTabDefinition'):
					self.root.addTabDefinition(obj)
				else:
					raise GuiXMLError("A Tab needs to be added to a TabbedArea widget!")
			else:
				self.root.addChild( obj )
		self.root = obj

	def endElement(self, name):
		self.indent = self.indent[:-4]
		if manager.debug: print(self.indent + "</%s>" % name)
		if self.stack.pop() in ('gui_element'):
			self.root = self.root.parent or self.root

def loadXML(filename_or_stream):
	"""
	Loads a PyChan XML file and generates a widget from it.

	@param filename_or_stream: A filename or a file-like object (for example using StringIO).

	The XML format is very dynamic, in the sense, that the actual allowed tags and attributes
	depend on the PyChan code.

	So when a tag C{Button} is encountered, an instance of class Button will be generated,
	and added to the parent object.
	All attributes will then be parsed and then set in the following way:

	  - position,size,min_size,max_size,margins - These are assumed to be comma separated tuples
	    of integers.
	  - foreground_color,base_color,background_color - These are assumed to be triples or quadruples of comma
	    separated integers. (triples: r,g,b; quadruples: r,g,b,a)
	  - border_size,padding - These are assumed to be simple integers.

	All other attributes are set verbatim as strings on the generated instance.
	In case a Widget does not accept an attribute to be set or the attribute can not be parsed
	correctly, the function will raise a GuiXMLError.

	In short::
		<VBox>
			<Button text="X" min_size="20,20" base_color="255,0,0" border_size="2" />
		</VBox>

	This result in the following code executed::

		vbox = VBox(parent=None)
		button = Button(parent=vbox)
		button.text = "X"
		button.min_size = (20,20)
		button.base_color = (255,0,0)
		button.border_size = 2
		vbox.add( button )
	"""
	from xml.sax import parse
	loader = _GuiLoader()
	parse(filename_or_stream,loader)
	return loader.root

def setupModalExecution(mainLoop,breakFromMainLoop):
	"""
	Setup the synchronous dialog execution feature.

	You can enable synchronous dialog execution by
	passing to functions to this function.

	@param mainLoop: Function - This is regarded as the applications
	main loop, which should be able to be called recursively.
	It should not take no arguments and return the argument
	passed to the second function (breakFromMainLoop).

	@param breakFromMainLoop: Function -This function should cause the
	first function to finish and return the passed argument.

	With these to functions dialogs can be executed synchronously.
	See L{widgets.Widget.execute}.
	"""
	if not manager:
		raise InitializationError("PyChan is not initialized yet.")
	manager.setupModalExecution(mainLoop,breakFromMainLoop)

def setUnicodePolicy(*policy):
	"""
	Set the unicode error handling policy.
	
	Possible options are:
	 - 'strict' meaning that encoding errors raise a UnicodeEncodeError.
	 - 'ignore' all encoding errors will be silently ignored.
	 - 'replace' all errors are replaced by the next argument.

	For further information look at the python documentation,
	especially C{codecs.register_error}.

	Example::
		pychan.setUnicodePolicy('replace','?')
	"""
	if not manager:
		raise InitializationError("PyChan is not initialized yet.")
	manager.unicodePolicy = policy






# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from builtins import map
from builtins import range
from builtins import object
from fife.extensions.pychan.attrs import IntAttr


AlignTop, AlignBottom, AlignLeft, AlignRight, AlignCenter = list(range(5))

def isLayouted(widget):
	return widget.real_widget.isLayouted()


# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from builtins import str
from fife import fife
from fife.extensions.serializers import ET

def loadXMLAnimation(engine, filename):
	f = engine.getVFS().open(filename)
	f.thisown = 1

	imgMgr = engine.getImageManager()
	aniMgr = engine.getAnimationManager()
	
	tree = ET.parse(f)
	node = tree.getroot()

	ani_id = node.get('id')
	if not ani_id:
		ani_id = filename

	if aniMgr.exists(ani_id):
		animation = aniMgr.getPtr(str(ani_id))
	else:
		animation = aniMgr.create(str(ani_id))
	
	common_width = int(node.get('width', 0))
	common_height = int(node.get('height', 0))
	common_frame_delay = int(node.get('delay', 0))
	common_x_offset = int(node.get('x_offset', 0))
	common_y_offset = int(node.get('y_offset', 0))
	animation.setActionFrame(int(node.get('action_frame', 0)))

	frames = node.findall('frame')
	if not frames:
		raise InvalidFormat('animation without <frame>s')

	atlas = node.get('atlas')

	if atlas:
		path = filename.split('/')
		path.pop()
		path.append(str(atlas))
		atlas_file = '/'.join(path)
		if imgMgr.exists(atlas_file):
			atlas_img = imgMgr.get(str(atlas_file))
		else:
			atlas_img = imgMgr.create(str(atlas_file))
		# parse atlas animation format 2 (e.g. cursor)
		for frame in frames:
			source = frame.get('source')
			if not source:
				raise InvalidFormat('animation without <frame>s')

			frame_x_pos = int(frame.get('xpos', 0))
			frame_y_pos = int(frame.get('ypos', 0))
			frame_x_offset = int(frame.get('x_offset', common_x_offset))
			frame_y_offset = int(frame.get('y_offset', common_y_offset))
			frame_delay = int(frame.get('delay', common_frame_delay))
			frame_width = int(frame.get('width', common_width))
			frame_height = int(frame.get('height', common_height))

			# xml paths are relative to the directory of the file they're used in.
			path = filename.split('/')
			path.pop()
			path.append(str(source))

			frame_file = '/'.join(path)
			frame_file = atlas,":",frame_file

			if imgMgr.exists(str(frame_file)):
				frame_img = imgMgr.get(str(frame_file))
			else:
				frame_img = imgMgr.create(str(frame_file))
				region = fife.Rect(frame_x_pos, frame_y_pos, frame_width, frame_height)
				frame_img.useSharedImage(atlas_img, region)
				frame_img.setXShift(frame_x_offset)
				frame_img.setYShift(frame_y_offset)

			animation.addFrame(frame_img, frame_delay)
	else:
		# parse single images
		for frame in frames:
			source = frame.get('source')
			if not source:
				raise InvalidFormat('animation without <frame>s')

			frame_x_offset = int(frame.get('x_offset', common_x_offset))
			frame_y_offset = int(frame.get('y_offset', common_y_offset))
			frame_delay = int(frame.get('delay', common_frame_delay))

			# xml paths are relative to the directory of the file they're used in.
			path = filename.split('/')
			path.pop()
			path.append(str(source))

			image_file = '/'.join(path)

			img = imgMgr.create(image_file)
			img.setXShift(frame_x_offset)
			img.setYShift(frame_y_offset)
			
			animation.addFrame(img, frame_delay)
			
#	animation.thisown = 0
	return animation



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from builtins import str
from builtins import map
from builtins import range
from fife import fifechan

from fife.extensions.pychan.attrs import Attr, UnicodeAttr

from .common import text2gui, gui2text
from .widget import Widget


class TextBox(Widget):
	"""
	An editable B{multiline} text edit widget.

	New Attributes
	==============

	  - text: The text in the TextBox.
	  - filename: A write-only attribute - assigning a filename will cause the widget to load it's text from it.

	Data
	====
	The text can be read and set via L{distributeData} and L{collectData}.
	"""

	ATTRIBUTES = Widget.ATTRIBUTES + [ UnicodeAttr('text'),
									   Attr('filename')
									 ]
	DEFAULT_HEXPAND = True
	DEFAULT_VEXPAND = True
	DEFAULT_MARGINS = 0, 0
	DEFAULT_PADDING = 0
	DEFAULT_TEXT = u""
	DEFAULT_FILENAME = ""

	def __init__(self,
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 text = None,
				 filename = None):

		self.real_widget = fifechan.TextBox()
		self.text = text or self.DEFAULT_TEXT
		self.filename = filename or self.DEFAULT_FILENAME
		super(TextBox,self).__init__(parent=parent,
									 name=name,
									 size=size,
									 min_size=min_size,
									 max_size=max_size,
									 fixed_size=fixed_size,
									 margins=margins,
									 padding=padding,
									 helptext=helptext,
									 position=position,
									 style=style,
									 hexpand=hexpand,
									 vexpand=vexpand,
									 font=font,
									 base_color=base_color,
									 background_color=background_color,
									 foreground_color=foreground_color,
									 selection_color=selection_color,
									 border_color=border_color,
									 outline_color=outline_color,
									 border_size=border_size,
									 outline_size=outline_size,
									 position_technique=position_technique,
									 is_focusable=is_focusable,
									 comment=comment)

		# Prepare Data collection framework
		self.accepts_data = True
		self.accepts_initial_data = True # Make sense in a way ...
		self._realSetInitialData = self._setText
		self._realSetData = self._setText
		self._realGetData = self._getText

	def clone(self, prefix):
		textboxClone = TextBox(None,
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size,
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext,
					self.position,
					self.style,
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.text,
					self.filename)
		return textboxClone


	def _getFileName(self): return self._filename
	def _loadFromFile(self,filename):
		self._filename = filename
		if not filename: return
		try:
			try:
				self.text = str(open(filename).read(),"utf8")
			except TypeError:
				self.text = open(filename).read()
		except Exception as e:
			self.text = str(e)
	filename = property(_getFileName, _loadFromFile)

	def _getText(self): return gui2text(self.real_widget.getText())
	def _setText(self,text): self.real_widget.setText(text2gui(text))
	text = property(_getText,_setText)

	def _setOpaque(self,opaque): self.real_widget.setOpaque(opaque)
	def _getOpaque(self): return self.real_widget.isOpaque()
	opaque = property(_getOpaque,_setOpaque)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from builtins import str
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import Attr,UnicodeAttr, PointAttr,BoolAttr,IntAttr

from .common import get_manager, gui2text, text2gui
from .containers import Container
from .widget import Widget


class Tab(Container):
	"""
	A tab widget.
	It can be used as a container.
	"""

	ATTRIBUTES = Container.ATTRIBUTES + [ UnicodeAttr('content_name'),
										]
	DEFAULT_HEXPAND = True
	DEFAULT_VEXPAND = False
	DEFAULT_MARGINS = 0, 0
	DEFAULT_PADDING = 0
	
	def __init__(self, 
				 parent = None, 
				 name = None,
				 size = None,
				 min_size = None, 
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None, 
				 position = None, 
				 style = None, 
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 _real_widget = None):

		if _real_widget is None: _real_widget = fifechan.Tab()
		# name of the widget that is associated with this tab
		self.content_name = None
		
		super(Tab,self).__init__(parent=parent,
								 name=name,
								 size=size,
								 min_size=min_size,
								 max_size=max_size,
								 fixed_size=fixed_size,
								 margins=margins,
								 padding=padding,
								 helptext=helptext,
								 position=position,
								 style=style,
								 hexpand=hexpand,
								 vexpand=vexpand,
								 font=font,
								 base_color=base_color,
								 background_color=background_color,
								 foreground_color=foreground_color,
								 selection_color=selection_color,
								 border_color=border_color,
								 outline_color=outline_color,
								 border_size=border_size,
								 outline_size=outline_size,
								 position_technique=position_technique,
								 is_focusable=is_focusable,
								 comment=comment,
								 background_image=background_image,
								 opaque=opaque,
								 _real_widget=_real_widget)
		
	def addTabToWidget(self, widget):
		if widget.tab is self: return
		if widget.tab is not None:
			self.removeTabFromWidget(widget)
		widget.tab = self

	def removeTabFromWidget(self, widget):
		if widget.tab == self:
			widget.tab = None

	def clone(self, prefix):
		tabClone = Tab(None, 
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size, 
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext, 
					self.position, 
					self.style, 
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.background_image,
					self.opaque)
					
		tabClone.addChildren(self._cloneChildren(prefix))
					
		return tabClone


class TabbedArea(Container):
	"""
	This is the tabbed area class.

	New Attributes
	==============


	"""
	ATTRIBUTES = Container.ATTRIBUTES + [ IntAttr('select_tab_index'),
										  UnicodeAttr('select_tab'),
										]

	DEFAULT_OPAQUE = False
	DEFAULT_MARGINS = 0, 0
	DEFAULT_PADDING = 0
	DEFAULT_LAYOUT = 'Horizontal'
	DEFAULT_POSITION_TECHNIQUE = "automatic"
	
	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 _real_widget = None,
				 select_tab_index = None,
				 select_tab = None):

		if _real_widget is None: _real_widget = fifechan.TabbedArea()
		
		super(TabbedArea,self).__init__(parent=parent,
										name=name,
										size=size,
										min_size=min_size,
										max_size=max_size,
										fixed_size=fixed_size,
										margins=margins,
										padding=padding,
										helptext=helptext,
										position=position,
										style=style,
										hexpand=hexpand,
										vexpand=vexpand,
										font=font,
										base_color=base_color,
										background_color=background_color,
										foreground_color=foreground_color,
										selection_color=selection_color,
										border_color=border_color,
										outline_color=outline_color,
										border_size=border_size,
										outline_size=outline_size,
										position_technique=position_technique,
										is_focusable=is_focusable,
										comment=comment,
										background_image=background_image,
										opaque=opaque,
										_real_widget=_real_widget)

		self.tab_definitions = {}
		# only use one
		if select_tab_index is not None: self.select_tab_index = select_tab_index
		elif select_tab is not None: self.select_tab = select_tab
		
	def clone(self, prefix):
		tabbedareaClone = TabbedArea(None, 
						self._createNameWithPrefix(prefix),
						self.size,
						self.min_size,
						self.max_size,
						self.fixed_size,
						self.margins,
						self.padding,
						self.helptext,
						self.position,
						self.style,
						self.hexpand,
						self.vexpand,
						self.font,
						self.base_color,
						self.background_color,
						self.foreground_color,
						self.selection_color,
						self.border_color,
						self.outline_color,
						self.border_size,
						self.outline_size,
						self.position_technique,
						self.is_focusable,
						self.comment,
						self.background_image,
						self.opaque,
						select_tab_index)
			
		tabbedareaClone.addChilds(self._cloneChildren(prefix))
			
		return tabbedareaClone
		

	def addChild(self, widget):
		"""
		Adds a child widget to the container.
		
		"""

		# if the widget have no tabwidget, we try to find one
		if widget.tab is None:
			tab = self.getTabDefinition(widget)
			tab.addTabToWidget(widget)
			
		widget.parent = self

		if widget.max_size[0] > self.max_size[0] or widget.max_size[1] > self.max_size[1]:
			widget.max_size = self.max_size

		self.children.append(widget)

		# add real tab and real widget
		self.real_widget.addTab(widget.tab.real_widget, widget.real_widget)
		# add all to the manager
		def _add(added_widget):
			if not added_widget._added:
				get_manager().addWidget(added_widget)
			if added_widget._top_added:
				get_manager().removeTopWidget(added_widget)
		widget.deepApply(_add)

	def removeChild(self, widget):
		if not widget in self.children:
			raise RuntimeError("%s does not have %s as direct tab widget." % (str(self),str(widget)))

		# remove all from the manager
		def _remove(removed_widget):
			if removed_widget._added:
				get_manager().removeWidget(removed_widget)
			if removed_widget._top_added:
				get_manager().removeTopWidget(removed_widget)
		widget.deepApply(_remove)

		i = self.children.index(widget)
		self.real_widget.removeTabWithIndex(i)
		self.children.remove(widget)
		widget.parent = None

	def addTabDefinition(self, widget):
		# Only needed because of XML loading
		# The tab have the content_name that should be equal to
		# the name of the contained widget
		self.tab_definitions[widget.content_name] = widget

	def getTabDefinition(self, widget):
		# Only needed because of XML loading
		# Check if we have a tab that is associated with the widget
		name = widget.name
		if name in self.tab_definitions:
			widget = self.tab_definitions[name]
			del self.tab_definitions[name]
			return widget
		else:
			raise RuntimeError("%s does not have %s in the tab definitions." % (str(self),str(name)))
		
	def getNumberOfTabs(self):
		return self.real_widget.getNumberOfTabs()

	def isTabSelected(self, widget):
		i = self.children.index(widget)
		return self.real_widget.isTabSelected(i)
	
	def _setSelectedTab(self, widget):
		if isinstance(widget, str):
			widget = self.findChild(name=widget)
		if widget is not None:
			i = self.children.index(widget)
			self.real_widget.setSelectedTab(i)
	def _getSelectedTab(self):
		i = self.real_widget.getSelectedTabIndex()
		return self.children[i]
	select_tab = property(_getSelectedTab, _setSelectedTab)
	
	def _setSelectedTabIndex(self, index):
		self.real_widget.setSelectedTab(index)
	def _getSelectedTabIndex(self):
		return self.real_widget.getSelectedTabIndex()
	select_tab_index = property(_getSelectedTabIndex, _setSelectedTabIndex)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fifechan

from fife.extensions.pychan.attrs import BoolAttr, ColorAttr, IntAttr, PointAttr
from fife.extensions.pychan.properties import ColorProperty

from .widget import Widget


class BarGraph(Widget):
	""" A bar graph widget

	New Attributes
	==============

	  - bar_position: x and y coordinate
	  - bar_width': int: default 10
	  - bar_height: int: default 10
	  - opaque: bool: default False
	"""

	ATTRIBUTES = Widget.ATTRIBUTES + [ PointAttr('bar_position'),
									   IntAttr('bar_width'),
									   IntAttr('bar_height'),
									   BoolAttr('opaque')
									 ]
	DEFAULT_HEXPAND = False
	DEFAULT_VEXPAND = False

	DEFAULT_OPAQUE = False
	DEFAULT_BAR_POSITION = 0,0
	DEFAULT_BAR_WIDTH = 10
	DEFAULT_BAR_HEIGHT = 10

	def __init__(self,
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 opaque = None,
				 bar_position = None,
				 bar_width = None,
				 bar_height = None):

		self.real_widget = fifechan.BarGraph()
		self.opaque = self.DEFAULT_OPAQUE
		self.bar_positon = self.DEFAULT_BAR_POSITION
		self.bar_width = self.DEFAULT_BAR_WIDTH
		self.bar_height = self.DEFAULT_BAR_HEIGHT

		super(BarGraph, self).__init__(parent=parent,
									   name=name,
									   size=size,
									   min_size=min_size,
									   max_size=max_size,
									   fixed_size=fixed_size,
									   margins=margins,
									   padding=padding,
									   helptext=helptext,
									   position=position,
									   style=style,
									   hexpand=hexpand,
									   vexpand=vexpand,
									   font=font,
									   base_color=base_color,
									   background_color=background_color,
									   foreground_color=foreground_color,
									   selection_color=selection_color,
									   border_color=border_color,
									   outline_color=outline_color,
									   border_size=border_size,
									   outline_size=outline_size,
									   position_technique=position_technique,
									   is_focusable=is_focusable,
									   comment=comment)

		if opaque is not None: self.opaque = opaque
		if bar_position is not None: self.bar_position = bar_position
		if bar_width is not None: self.bar_width = bar_width
		if bar_height is not None: self.bar_height = bar_height


	def clone(self, prefix):
		barGraphClone = BarGraph(None,
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size,
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext,
					self.position,
					self.style,
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.opaque,
					self.bar_positon,
					self.bar_width,
					self.bar_height)
		return barGraphClone

	def _setOpaque(self, opaque): self.real_widget.setOpaque(opaque)
	def _getOpaque(self): return self.real_widget.isOpaque()
	opaque = property(_getOpaque, _setOpaque)

	def _setBarPosition(self, bar_position): self.real_widget.setBarPosition(bar_position[0], bar_position[1])
	def _getBarPosition(self): return (self.real_widget.getBarX(), self.real_widget.getBarY())
	bar_position = property(_getBarPosition, _setBarPosition)

	def _setBarWidth(self, bar_width): self.real_widget.setBarWidth(bar_width)
	def _getBarWidth(self): return self.real_widget.getBarWidth()
	bar_width = property(_getBarWidth, _setBarWidth)
	
	def _setBarHeight(self, bar_height): self.real_widget.setBarHeight(bar_height)
	def _getBarHeight(self): return self.real_widget.getBarHeight()
	bar_height = property(_getBarHeight, _setBarHeight)




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Automatic widget positioning
============================

You can use the C{position_technique} attribute
on top level widgets which can also be set from xml.

For direct use call L{placeWidget}.
"""
from __future__ import absolute_import

from .internal import screen_width, screen_height
from .exceptions import PyChanException

EXPLICIT = "explicit"
AUTOMATIC = "automatic"

TOP = "top"
LEFT = "left"
RIGHT = "right"
CENTER = "center"
BOTTOM = "bottom"


def _splicePosition(p):
	if "+" in p:
		technique,delta = p.split("+")
	elif "-" in p:
		technique,delta = p.split("-")
		delta = '-' + delta
	else:
		technique,delta = p,0
	delta = int(delta)
	return technique,delta

def _parsePosition(position):
	try:
		if position == AUTOMATIC:
			position = "center+0:center+0"

		x_pos,y_pos = position.split(":")
		x_pos, x_delta = _splicePosition(x_pos)
		y_pos, y_delta = _splicePosition(y_pos)

		if x_pos not in [EXPLICIT,LEFT,CENTER,RIGHT]:
			raise ""
		if y_pos not in [EXPLICIT,TOP,CENTER,BOTTOM]:
			raise ""
	except:
		raise PyChanException("Malformed position definition: " + repr(position))
	return x_pos,x_delta,y_pos,y_delta

def placeWidget(widget,position):
	"""
	Place a widget according to a string defining relative coordinates to screen borders.

	The position definition has to be of the form: C{"<x_pos><x_delta>:<y_pos><y_delta>"}

	C{<x_pos>} may be one of:
          - left
          - right
          - center
          - explicit

	C{<y_pos>} may be one of:
          - top
          - bottom
          - center
          - explicit

        C{explicit} means that the widgets x or y position will not be touched. The rest should be
        self explanatory.

	C{<x_delta>} and C{<y_delta>} must be of the form: +pixel_number or -pixel_number. Or completely
        omitted. Note that the sign has to be there for for positive deltas, too.

	For brevity two shortcuts exist:
          - "explicit" -> "explicit:explicit"
          - "automatic" -> "center:center"

	A few examples::
          "right-20:top"
          "center:top+10"
          "center:center"

	@param widget: The PyChan widget.
	@param position: A position definition.

	If the position cannot be parsed a L{PyChanException} is thrown.

	"""
	if position == EXPLICIT:
		return
	x_pos,x_delta,y_pos,y_delta = _parsePosition(position)

	x,y = widget.position
	w,h = widget.size

	if x_pos == CENTER:
		x = (screen_width() - w) // 2 + x_delta

	if y_pos == CENTER:
		y = (screen_height() - h) // 2 + y_delta

	if x_pos == LEFT:
		x = x_delta

	if y_pos == TOP:
		y = y_delta

	if x_pos == RIGHT:
		x = screen_width() - w + x_delta

	if y_pos == BOTTOM:
		y = screen_height() - h + y_delta

	widget.position = x,y






# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fife
from fife import fifechan

from fife.extensions.pychan.attrs import BoolAttr, MixedListAttr
from .common import get_manager

from .containers import Window

class ResizableWindow(Window):
	"""
	This is a window class that can be resizable.

	New Attributes
	==============

	  - resizable: Allows resizing with the mouse. If disabled all the side parameters are ignored.
	  - top_resizable: If true, the window can be resized at the top side.
	  - right_resizable: If true, the window can be resized at the right side.
	  - bottom_resizable: If true, the window can be resized at the bottom side.
	  - left_resizable: If true, the window can be resized at the left side.
	  - shove: That allows pull and push in case the min/max size is reached.
	  - cursors: List that contains the cursor definitions. Left, Right, Top, Bottom, LeftTop, RightTop, LeftBottom, RightBottom
	"""

	DEFAULT_RESIZABLE = True
	DEFAULT_TOP_RESIZABLE = True
	DEFAULT_RIGHT_RESIZABLE = True
	DEFAULT_BOTTOM_RESIZABLE = True
	DEFAULT_LEFT_RESIZABLE = True
	DEFAULT_SHOVE = False
	
	
	ATTRIBUTES = Window.ATTRIBUTES + [ BoolAttr('resizable'),
									   BoolAttr('top_resizable'),
									   BoolAttr('right_resizable'),
									   BoolAttr('bottom_resizable'),
									   BoolAttr('left_resizable'),
									   BoolAttr('shove'),
									   MixedListAttr('cursors')
									 ]

	DIRECTION_LIST = [ fifechan.ResizableWindow.CURSOR_DIRECTION_L,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_R,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_T,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_B,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_LT,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_RT,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_LB,
					   fifechan.ResizableWindow.CURSOR_DIRECTION_RB
					 ]

	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 background_image = None,
				 opaque = None,
				 layout = None,
				 spacing = None,
				 uniform_size = None,
				 _real_widget = None,
				 title = None,
				 titlebar_height = None,
				 movable = None,
				 resizable = None,
				 top_resizable = None,
				 right_resizable = None,
				 bottom_resizable = None,
				 left_resizable = None,
				 shove = None,
				 cursors = None):
		
		if _real_widget is None: _real_widget = fifechan.ResizableWindow()
			
		super(ResizableWindow,self).__init__(parent=parent,
											 name=name,
											 size=size,
											 min_size=min_size,
											 max_size=max_size,
											 fixed_size=fixed_size,
											 margins=margins,
											 padding=padding,
											 helptext=helptext,
											 position=position,
											 style=style,
											 hexpand=hexpand,
											 vexpand=vexpand,
											 font=font,
											 base_color=base_color,
											 background_color=background_color,
											 foreground_color=foreground_color,
											 selection_color=selection_color,
											 border_color=border_color,
											 outline_color=outline_color,
											 border_size=border_size,
											 outline_size=outline_size,
											 position_technique=position_technique,
											 is_focusable=is_focusable,
											 comment=comment,
											 background_image=background_image,
											 opaque=opaque,
											 layout=layout,
											 spacing=spacing,
											 uniform_size=uniform_size,
											 _real_widget=_real_widget,
											 title=title,
											 titlebar_height=titlebar_height,
											 movable=movable)

		if resizable is not None: self.resizable = resizable
		else: self.resizable = self.DEFAULT_RESIZABLE

		if top_resizable is not None: self.top_resizable = top_resizable
		else: self.top_resizable = self.DEFAULT_TOP_RESIZABLE

		if right_resizable is not None: self.right_resizable = right_resizable
		else: self.right_resizable = self.DEFAULT_RIGHT_RESIZABLE

		if bottom_resizable is not None: self.bottom_resizable = bottom_resizable
		else: self.bottom_resizable = self.DEFAULT_BOTTOM_RESIZABLE

		if left_resizable is not None: self.left_resizable = left_resizable
		else: self.left_resizable = self.DEFAULT_LEFT_RESIZABLE

		if shove is not None: self.shove = shove
		else: self.shove = self.DEFAULT_SHOVE

		if cursors is not None: self.cursors = cursors

	def clone(self, prefix):
		windowClone = ResizableWindow(None, 
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size, 
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext, 
					self.position, 
					self.style, 
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.background_image,
					self.opaque,
					self.layout,
					self.spacing,
					self.uniform_size,
					None,
					self.title,
					self.titlebar_height,
					self.movable,
					self.resizable,
					self.top_resizable,
					self.right_resizable,
					self.bottom_resizable,
					self.left_resizable,
					self.shove,
					self.cursors)
		
		windowClone.addChildren(self._cloneChildren(prefix))		     
		return windowClone

	def _setResizable(self, resizable): self.real_widget.setResizable(resizable)
	def _getResizable(self): return self.real_widget.getResizable()
	resizable = property(_getResizable, _setResizable)

	def _setTopResizable(self, resizable): self.real_widget.setTopResizable(resizable)
	def _getTopResizable(self): return self.real_widget.getTopResizable()
	top_resizable = property(_getTopResizable, _setTopResizable)

	def _setRightResizable(self, resizable): self.real_widget.setRightResizable(resizable)
	def _getRightResizable(self): return self.real_widget.getRightResizable()
	right_resizable = property(_getRightResizable, _setRightResizable)

	def _setBottomResizable(self, resizable): self.real_widget.setBottomResizable(resizable)
	def _getBottomResizable(self): return self.real_widget.getBottomResizable()
	bottom_resizable = property(_getBottomResizable, _setBottomResizable)

	def _setLeftResizable(self, resizable): self.real_widget.setLeftResizable(resizable)
	def _getLeftResizable(self): return self.real_widget.getLeftResizable()
	left_resizable = property(_getLeftResizable, _setLeftResizable)

	def _setShove(self, shove): self.real_widget.setShove(shove)
	def _getShove(self): return self.real_widget.getShove()
	shove = property(_getShove, _setShove)

	def _setCursor(self, direction, cursor):
		self.real_widget.set(direction, cursor)
	def _getCursor(self, direction):
		type = self.real_widget.getType(direction)
		if type is fife.CURSOR_NATIVE:
			return self.real_widget.getId(direction)
		elif type is fife.CURSOR_IMAGE:
			return self.real_widget.getImage(direction)
		elif type is fife.CURSOR_Animation:
			return self.real_widget.getAnimation(direction)
	cursor = property(_getCursor, _setCursor)

	def _setCursors(self, cursors):
		if cursors is not None:
			i = 0
			for c in cursors:
				if isinstance(c, str):
					# we need a fife image
					c = get_manager().loadImage(c, gui=False)
				self.real_widget.set(self.DIRECTION_LIST[i], c)
				i += 1
	def _getCursors(self):
		cursors = []
		for d in self.DIRECTION_LIST:
			cursors.append(self.cursor(d))
		return cursors
	cursors = property(_getCursors, _setCursors)




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
from fife.extensions import pychan
from fife.extensions.pychan import loadXML
import fife.extensions.pychan.tools as tools
from .internal import get_manager, screen_width, screen_height
from io import StringIO

OK,YES,NO,CANCEL = True,True,False,None

def print_event(**kwargs):
	print(kwargs)

class XMLDialog(object):
	def __init__(self, xml, ok_field = None, cancel_field = None,initial_data={},data={}):
		self.gui = loadXML(xml)
		self.ok_field = ok_field
		self.cancel_field = cancel_field
		self.initial_data= initial_data
		self.data= data
		self.max_size=None
		self.min_size=None
#		self.gui.capture(print_event,"mouseEntered")

	def execute(self):
		self.gui.distributeInitialData(self.initial_data)
		self.gui.distributeData(self.data)

		screen_w, screen_h = screen_width(), screen_height()
		if self.max_size is None:
			self.max_size = screen_w // 2, screen_h // 3
		if self.min_size is None:
			self.min_size = screen_w // 2, screen_h // 4
		self.gui.max_size = self.max_size
		self.gui.min_size = self.min_size

		resultMap = {}
		if self.gui.findChild(name="okButton"):
			resultMap["okButton"] = OK

		if self.gui.findChild(name="cancelButton"):
			resultMap["cancelButton"] = CANCEL

		if self.gui.findChild(name="yesButton"):
			resultMap["noButton"] = NO

		if self.gui.findChild(name="yesButton"):
			resultMap["yesButton"] = YES

		ok = self.gui.execute(resultMap)
		if ok:
			return self.getOkResult()
		return self.getCancelResult()

	def getOkResult(self):
		if self.ok_field:
			return self.gui.collectData(self.ok_field)
		return True

	def getCancelResult(self):
		if self.cancel_field:
			return self.gui.collectData(self.cancel_field)
		return False

MESSAGE_BOX_XML = """\
<Window name="window" title="Message">
<ScrollArea>
<Label wrap_text="1" text="$MESSAGE" name="message" vexpand="True"/>
</ScrollArea>
<HBox>
<Spacer/><Button min_size="50,0" name="okButton" text="OK"/>
</HBox>
</Window>
"""

YESNO_BOX_XML = """\
<Window name="window" title="Question">
<ScrollArea>
<Label wrap_text="1" text="$MESSAGE" name="message" vexpand="True"/>
</ScrollArea>
<HBox>
<Spacer/>
<Button min_size="50,0" name="yesButton" text="Yes"/>
<Button min_size="50,0" name="noButton" text="No"/>
</HBox>
</Window>
"""

YESNOCANCEL_BOX_XML = """\
<Window name="window" title="Question">
<ScrollArea>
<Label wrap_text="1" text="$MESSAGE" name="message" vexpand="True"/>
</ScrollArea>
<HBox>
<Spacer/>
<Button min_width="50" name="yesButton" text="Yes"/>
<Button min_width="50" name="noButton" text="No"/>
<Button min_width="50" name="cancelButton" text="Cancel"/>
</HBox>
</Window>
"""

SELECT_BOX_XML = """\
<Window name="window" title="Select">
<Label wrap_text="1" text="$MESSAGE" name="message"/>
<ScrollArea>
<ListBox name="selection">
</ListBox>
</ScrollArea>
<HBox>
<Spacer/>
<Button min_width="50" name="okButton" text="Select"/>
<Button min_width="50" name="cancelButton" text="Cancel"/>
</HBox>
</Window>
"""

EXCEPTION_CATCHER_XML= """\
<Window name="window" title="An exception occurred - what now?">
  <VBox hexpand="True">
    <Label wrap_text="1" max_size="400,90000" text="$MESSAGE" name="message"/>
    <ScrollArea>
    <Label text="$MESSAGE" name="traceback"/>
    </ScrollArea>
  </VBox>
  <HBox>
    <Spacer/>
    <Button name="yesButton" text="Retry"/>
    <Button name="noButton" text="Ignore"/>
    <Button name="cancelButton" text="Reraise"/>
  </HBox>
</Window>
"""

def _make_text(message):
	if callable(message):
		message = message()
	if hasattr(message,"read"):
		message = message.read()
	return message

def message(message="",caption="Message"):
	text = _make_text(message)
	dialog = XMLDialog(StringIO(MESSAGE_BOX_XML))
	dialog.gui.findChild(name="message").max_width = screen_width() // 2 - 50
	dialog.gui.findChild(name="message").text = text
	dialog.gui.findChild(name="window").title = caption

	dialog.execute()

def yesNo(message="",caption="Message"):
	text = _make_text(message)
	dialog = XMLDialog(StringIO(YESNO_BOX_XML))
	dialog.gui.findChild(name="message").max_width = screen_width() // 2 - 50
	dialog.gui.findChild(name="message").text = text
	dialog.gui.findChild(name="window").title = caption

	return dialog.execute()

def yesNoCancel(message="",caption="Message"):
	text = _make_text(message)
	dialog = XMLDialog(StringIO(YESNOCANCEL_BOX_XML))
	dialog.gui.findChild(name="message").max_width = screen_width() // 2 - 50
	dialog.gui.findChild(name="message").text = text
	dialog.gui.findChild(name="window").title = caption

	return dialog.execute()

def select(message="",options=[],caption="Message"):
	text = _make_text(message)
	dialog = XMLDialog(StringIO(SELECT_BOX_XML))
	dialog.size = screen_width() // 3, (2 * screen_height()) // 3
	dialog.gui.findChild(name="message").max_width = screen_width() // 2 - 50
	dialog.gui.findChild(name="message").text = text
	dialog.gui.findChild(name="window").title = caption
	
	listbox = dialog.gui.findChild(name="selection")
	listbox.items = options
	if dialog.execute():
		return listbox.selected_item
	return None

def trace(f):
	import sys, traceback
	def new_f(*args,**kwargs):
		try:
			return pychan.tools.applyOnlySuitable(f,*args,**kwargs)

		except Exception as e:
			dialog = XMLDialog(StringIO(EXCEPTION_CATCHER_XML))
			dialog.gui.findChild(name="message").text = str(e)
			
			tb = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
			dialog.gui.findChild(name="traceback").text = "".join(tb)
			dialog.min_size = screen_width() // 2, (3 * screen_height()) // 4
			dialog.max_size = screen_width() // 2, (3 * screen_height()) // 4
			result = dialog.execute()
			if result == YES:
				return new_f(*args,**kwargs)
			elif result == NO:
				return
			raise
	return new_f




# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
PychanSettings
==================================

This module provides gui window that sets basic fife settings through
pychan.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
import os
from io import BytesIO, StringIO, StringIO

from fife.extensions.fife_settings import Setting
from fife.extensions.fife_settings import SettingEntry
from fife.extensions import pychan

SETTINGS_GUI_XML="""\
<Window name="Settings" title="Settings">
	<Label text="Settings menu!" />
	<HBox>
		<VBox>
			<Label text="Resolution:" />
			<Label text="Renderer:" />
		</VBox>
		<VBox min_size="120,60">
			<DropDown name="screen_resolution" min_size="120,0" />
			<DropDown name="render_backend" min_size="120,0" />
		</VBox>
	</HBox>
	<CheckBox name="enable_fullscreen" text="Use the full screen mode" />
	<CheckBox name="enable_sound" text="Enable sound" />
	<HBox>
		<Spacer />
		<Button name="cancelButton" text="Cancel" />
		<Button name="okButton" text="Ok" />
		<Button name="defaultButton" text="Defaults" />
	</HBox>
</Window>
"""

CHANGES_REQUIRE_RESTART="""\
<Window title="Changes require restart">
	<Label text="Some of your changes require you to restart." />
	<HBox>
		<Spacer />
		<Button name="closeButton" text="Ok" />
	</HBox>
</Window>
"""

FIFE_MODULE = "FIFE"

class FifePychanSettings(Setting):
	def __init__(self, app_name="", settings_file="", default_settings_file= "settings-dist.xml", settings_gui_xml="", changes_gui_xml="", copy_dist=True, serializer=None):
		
		super(FifePychanSettings, self).__init__(app_name, settings_file, default_settings_file, copy_dist, serializer)
		
		self._optionsDialog = None
		self._gui_style = "default"
		self.changesRequireRestart = False
		
		if settings_gui_xml == "":
			self._settings_gui_xml = SETTINGS_GUI_XML
		else:
			self._settings_gui_xml = settings_gui_xml
			
		if changes_gui_xml == "":
			self._changes_gui_xml = CHANGES_REQUIRE_RESTART
		else:
			self._changes_gui_xml = changes_gui_xml
		
	def createAndAddEntry(self, module, name, widgetname, applyfunction=None, initialdata=None, requiresrestart=False):
		""""
		@param module: The Setting module this Entry belongs to
		@type module: C{String}
		@param name: The Setting's name
		@type name: C{String}
		@param widgetname: The name of the widget that is used to change this
		setting
		@type widgetname: C{String}
		@param applyfunction: function that makes the changes when the Setting is
		saved
		@type applyfunction: C{function}
		@param initialdata: If the widget supports the setInitialData() function
		this can be used to set the initial data
		@type initialdata: C{String} or C{Boolean}
		@param requiresrestart: Whether or not the changing of this setting
		requires a restart
		@type requiresrestart: C{Boolean}
		"""
		entry = PychanSettingEntry(module, name, widgetname, applyfunction, initialdata, requiresrestart)
		self.addEntry(entry)
		

	def _initDefaultSettingEntries(self):
		"""Initializes the default fife setting entries. Not to be called from
		outside this class."""
		self.createAndAddEntry(FIFE_MODULE, "PlaySounds", "enable_sound",
		              requiresrestart=True)
		self.createAndAddEntry(FIFE_MODULE, "FullScreen", "enable_fullscreen",
		              requiresrestart=True)
		self.createAndAddEntry(FIFE_MODULE, "ScreenResolution", "screen_resolution", initialdata = self._resolutions,
		              requiresrestart=True)
		self.createAndAddEntry(FIFE_MODULE, "RenderBackend", "render_backend", initialdata = self._renderbackends,
		              requiresrestart=True)
		
	# sets valid resolution options in the settings->Resolution
	def setValidResolutions(self, options):
		if options:
			self._resolutions = options
		self.createAndAddEntry(FIFE_MODULE, "ScreenResolution", "screen_resolution", initialdata = self._resolutions,
		              requiresrestart=True)

		              
	def setGuiStyle(self, style):
		""" 
		Set a custom gui style used for the option dialog.
		@param style: Pychan style to be used
		@type style: C{string}      
		"""
		self._gui_style = style
		
	def showSettingsDialog(self):
		"""
		Shows the default settings dialog.
		"""
		self.changesRequireRestart = False
		self.isSetToDefault = False
		if not self._optionsDialog:
			self._loadSettingsDialog()
		
		self.fillWidgets()
		self._optionsDialog.show()
		
	def _loadSettingsDialog(self):
		"""
		Loads up the .xml with the default settings dialog.
		"""
		self._optionsDialog = self._loadWidget(self._settings_gui_xml)
		self._optionsDialog.stylize(self._gui_style)
		self._optionsDialog.mapEvents({
			'okButton' : self._applySettings,
			'cancelButton': self._optionsDialog.hide,
			'defaultButton': self.setDefaults
		})
		return self._optionsDialog
		
	def _loadWidget(self, widget):
		"""
		Loads a widget
		"""
		if os.path.isfile(self._settings_gui_xml):
			return pychan.loadXML(widget)
		else:
			try:
				return pychan.loadXML(BytesIO(widget))
			except TypeError:
				return pychan.loadXML(StringIO(widget))

	def fillWidgets(self):
		for module in self._entries.values():
			for entry in module.values():
				widget = self._optionsDialog.findChildByName(entry.settingwidgetname)
				
				"""
				little change to prevent crash from no settings
				in settings.xml file
				"""
				"""
				The checking of value for None is specially for the clients who use settings
				with different names under modules other than "FIFE" for which we have no
				default value to set. This will prevent the settings widget from crash
				"""
				value = self.get(entry.module, entry.name)
				
				if isinstance(entry.initialdata, list):
					try:
						value = entry.initialdata.index(value)
					except ValueError:
						raise ValueError("\"" + str(value) + "\" is not a valid value for " + entry.name + ". Valid options: " + str(entry.initialdata))
				elif isinstance(entry.initialdata, dict):
					try:
						value = list(entry.initialdata.keys()).index(value)
					except ValueError:
						raise ValueError("\"" + str(value) + "\" is not a valid value for " + entry.name + ". Valid options: " + str(list(entry.initialdata.keys())))
				entry.initializeWidget(widget, value)
	
	def _applySettings(self):
		"""
		Writes the settings file.  If a change requires a restart of the engine
		it notifies you with a small dialog box.
		"""
		for module in self._entries.values():
			for entry in module.values():
				widget = self._optionsDialog.findChildByName(entry.settingwidgetname)
				data = widget.getData()
				
				# If the data is a list we need to get the correct selected data
				# from the list. This is needed for e.g. dropdowns or listboxs
				if type(entry.initialdata) is list:
					data = entry.initialdata[data]
				elif isinstance(entry.initialdata, dict):
					data = list(entry.initialdata.keys())[data]

				# only take action if something really changed
				if data != self.get(entry.module, entry.name):
					self.set(entry.module, entry.name, data)
					entry.onApply(data)

					if entry.requiresrestart:
						self.changesRequireRestart = True

		super(FifePychanSettings, self).saveSettings()

		self._optionsDialog.hide()
		if self.changesRequireRestart:
			self._showChangeRequireRestartDialog()
			
	def _showChangeRequireRestartDialog(self):
		"""Shows a dialog that informes the user that a restart is required
		to perform the changes."""
		RestartDlg = self._loadWidget(self._changes_gui_xml)
		RestartDlg.stylize(self._gui_style)
		RestartDlg.mapEvents({ 'closeButton' : RestartDlg.hide })
		RestartDlg.show()
		
	def setDefaults(self):
		super(FifePychanSettings, self).setDefaults()
		
		#On startup the settings dialog is not yet initialized.  We dont
		#fill the widgets with data in that case.
		if self._optionsDialog:
			self.fillWidgets()
			
class PychanSettingEntry(SettingEntry):
	
	def __init__(self, module, name, widgetname, applyfunction=None, initialdata=None, requiresrestart=False):
		super(PychanSettingEntry, self).__init__(module, name, applyfunction, initialdata, requiresrestart)
		
		self._settingwidgetname = widgetname
		
	def initializeWidget(self, widget, currentValue):
		"""Initialize the widget with needed data"""
		if self._initialdata is not None:
			widget.setInitialData(self._initialdata)
		widget.setData(currentValue)
	
	def _getSettingWidgetName(self):
		return self._settingwidgetname

	def _setSettingWidgetName(self, settingwidgetname):
		self._settingwidgetname = settingwidgetname
		
	settingwidgetname = property(_getSettingWidgetName, _setSettingWidgetName)
	
	def __str__(self):
		return "SettingEntry: " +  self.name + " Module: " + self.module + " Widget: " + \
		       self.settingwidgetname + " requiresrestart: " + str(self.requiresrestart) + \
		       " initialdata: " + str(self.initialdata)

	



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Settings
==================================

This module provides a nice framework for loading and saving game settings.
It is by no means complete but it does provide a good starting point.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
import shutil
import os
from io import StringIO

from fife.extensions import fifelog
from fife.extensions.fife_utils import getUserDataDirectory
from fife.extensions.serializers.simplexml import SimpleXMLSerializer

FIFE_MODULE = "FIFE"

class Setting(object):
	"""
	This class manages loading and saving of game settings.

	Usage::
		from fife.extensions.fife_settings import Setting
		settings = Setting(app_name="myapp")
		screen_width = settings.get("FIFE", "ScreenWidth", 1024)
		screen_height = settings.get("FIFE", "ScreenHeight", 768)
	"""

	def __init__(self, app_name="", settings_file="", default_settings_file= "settings-dist.xml", copy_dist=True, serializer=None):
		r"""
		Initializes the Setting object.

		@param app_name: The applications name.  If this parameter is provided
		alone it will try to read the settings file from the users home directory.
		In windows this will be	something like:	C:\Documents and Settings\user\Application Data\fife
		@type app_name: C{string}
		@param settings_file: The name of the settings file.  If this parameter is
		provided it will look for the setting file as you specify it, first looking
		in the working directory.  It will NOT look in the users home directory.
		@type settings_file: C{string}
		@param default_settings_file: The name of the default settings file.  If the settings_file
		does not exist this file will be copied into the place of the settings_file.  This file
		must exist in the root directory of your project!
		@type default_settings_file: C{string}
		@param copy_dist: Copies the default settings file to the settings_file location.  If
		this is False it will create a new empty setting file.
		@param serializer: Overrides the default XML serializer
		@type serializer: C{SimpleSerializer}

		"""
		self._app_name = app_name
		self._settings_file = settings_file
		self._default_settings_file = default_settings_file

		# Holds SettingEntries
		self._entries = {}

		if self._settings_file == "":
			self._settings_file = "settings.xml"
			self._appdata = getUserDataDirectory("fife", self._app_name)
		else:
			self._appdata = os.path.dirname(self._settings_file)
			self._settings_file = os.path.basename(self._settings_file)

		if not os.path.exists(os.path.join(self._appdata, self._settings_file)):
			if os.path.exists(self._default_settings_file) and copy_dist:
				shutil.copyfile(self._default_settings_file, os.path.join(self._appdata, self._settings_file))

		# valid values possible for the engineSettings
		self._validSetting = {}
		self._validSetting['FIFE'] = {
			'FullScreen':[True,False], 'RefreshRate':[0,200], 'Display':[0,9], 'VSync':[True,False], 'PychanDebug':[True,False]
			, 'ProfilingOn':[True,False], 'SDLRemoveFakeAlpha':[True,False], 'GLCompressImages':[False,True], 'GLUseFramebuffer':[False,True], 'GLUseNPOT':[False,True],
			'GLUseMipmapping':[False,True], 'GLTextureFiltering':['None', 'Bilinear', 'Trilinear', 'Anisotropic'], 'GLUseMonochrome':[False,True],
			'GLUseDepthBuffer':[False,True], 'GLAlphaTestValue':[0.0,1.0],
			'RenderBackend':['OpenGL', 'SDL'],
			'ScreenResolution':['640x480', '800x600', '1024x600', '1024x768', '1280x768',
								'1280x800', '1280x960', '1280x1024', '1366x768', '1440x900',
								'1600x900', '1600x1200', '1680x1050', '1920x1080', '1920x1200'],
			'BitsPerPixel':[0,16,24,32],
			'InitialVolume':[0.0,10.0], 'WindowTitle':"", 'WindowIcon':"", 'Font':"",
			'FontGlyphs':"", 'DefaultFontSize':"", 'Lighting':[0,1],
			'ColorKeyEnabled':[True,False], 'ColorKey':['a','b','c'], 'VideoDriver':"", 'RenderDriver':"",
			'PlaySounds':[True,False], 'LogToFile':[True,False],
			'LogToPrompt':[True,False], 'LogLevelFilter':[0,1,2,3],
			'LogModules':['all', 'controller','script','video','audio','loaders','vfs','pool','view','model','metamodel','event_channel','xml'],
			'FrameLimitEnabled':[True,False], 'FrameLimit':[0], 'MouseSensitivity':[0.0], 'MouseAcceleration':[True,False],
			'NativeImageCursor':[True,False], 'JoystickSupport':[True, False]
			}

		glyphDft = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&amp;`'*#=[]\\\""

		# we at this point assume default values are final values for engineSettings
		self._defaultSetting = {}
		self._defaultSetting['FIFE'] = {
			'FullScreen':False, 'RefreshRate':60, 'Display':0, 'VSync':False, 'PychanDebug':False,
			'ProfilingOn':False, 'SDLRemoveFakeAlpha':False, 'GLCompressImages':False, 'GLUseFramebuffer':True, 'GLUseNPOT':True,
			'GLUseMipmapping':False, 'GLTextureFiltering':'None', 'GLUseMonochrome':False, 'GLUseDepthBuffer':False, 'GLAlphaTestValue':0.3,
			'RenderBackend':'OpenGL', 'ScreenResolution':"1024x768", 'BitsPerPixel':0,
			'InitialVolume':5.0, 'WindowTitle':"", 'WindowIcon':"", 'Font':"",
			'FontGlyphs':glyphDft, 'DefaultFontSize':12, 'Lighting':0,
			'ColorKeyEnabled':False, 'ColorKey':[255,0,255], 'VideoDriver':"", 'RenderDriver':"",
			'PlaySounds':True, 'LogToFile':False,
			'LogToPrompt':False,'LogLevelFilter':0,
			'LogModules':['controller','script'],
			'FrameLimitEnabled':False, 'FrameLimit':60,
			'MouseSensitivity':0.0,
			'MouseAcceleration':False,
			'NativeImageCursor':False,
                        'JoystickSupport':False
			}

		# has the settings file been read
		self._readSettingsCompleted = {}

		# the global dictionary from which we will read after self._readSettingsCompleted is True
		self._settingsFromFile = {}

		# the logger needed to write in log file. It will be initialized in this file when self.getSettings()
		# method is called by logger
		self._logger = None

		#default settings
		self._resolutions = self._validSetting['FIFE']['ScreenResolution']
		self._renderbackends = self._validSetting['FIFE']['RenderBackend']
		self._lightingmodels = self._validSetting['FIFE']['Lighting']

		#Used to stylize the options gui
		self._gui_style = "default"

		#Initialize the serializer
		if serializer:
			self._serializer = serializer
		else:
			self._serializer = SimpleXMLSerializer()

		self.initSerializer()

		# if there's no FIFE module assume the settings file is broken
		# and replace with default settings file
		if "FIFE" not in self._serializer.getModuleNameList():
			self.setDefaults()

		# Get all modules and initialize reading of them from xml file as false
		self._allModules = self._serializer.getModuleNameList()
		# print("All Module Names:",self._allModules)
		for module in self._allModules:
			self._readSettingsCompleted[module] = False

		self._initDefaultSettingEntries()

		#self.setOneSetting('FIFE','Font','fonts/FreeSans.ttf',False)

		#print self.getSettingsFromFile('unknownhorizons')

	# set all Settings in either validSetting or defaultSetting
	def setAllSettings(self,module,settings,validSetting = True):
		if validSetting:
			self._validSetting[module] = settings
		else:
			self._defaultSetting[module] = settings

	# set an entry in the validSetting or defaultSetting dictionary
	def setOneSetting(self,module,name,value,validSetting = True):
		if validSetting:
			self._validSetting[module][name] = value
		else:
			self._defaultSetting[module][name] = value

	# get all the Settings(either validSetting or defaultSetting)
	def getAllSettings(self,module,validSetting = True):
		if validSetting:
			return self._validSetting[module]
		else:
			return self._defaultSetting[module]

	# get an entry from either validSetting or defaultSetting
	def getOneSetting(self,module,name,validSetting = True):
		if validSetting:
			return self._validSetting[module][name]
		else:
			return self._defaultSetting[module][name]

	# sets valid resolution options in the settings->Resolution
	def setValidResolutions(self, options):
		if options:
			self._resolutions = options
		self.createAndAddEntry(FIFE_MODULE, "ScreenResolution", initialdata = self._resolutions,
		              requiresrestart=True)

	def initSerializer(self):
		self._serializer.load(os.path.join(self._appdata, self._settings_file))

	def _initDefaultSettingEntries(self):
		"""Initializes the default fife setting entries. Not to be called from
		outside this class."""
		self.createAndAddEntry(FIFE_MODULE, "PlaySounds", requiresrestart=True)

		self.createAndAddEntry(FIFE_MODULE, "FullScreen", requiresrestart=True)

		self.createAndAddEntry(FIFE_MODULE, "ScreenResolution", initialdata = self._resolutions, requiresrestart=True)

		self.createAndAddEntry(FIFE_MODULE, "RenderBackend", initialdata = self._renderbackends, requiresrestart=True)

	def createAndAddEntry(self, module, name, applyfunction=None, initialdata=None, requiresrestart=False):
		""""
		@param module: The Setting module this Entry belongs to
		@type module: C{String}
		@param name: The Setting's name
		@type name: C{String}
		@param applyfunction: function that makes the changes when the Setting is
		saved
		@type applyfunction: C{function}
		@param initialdata: If the widget supports the setInitialData() function
		this can be used to set the initial data
		@type initialdata: C{String} or C{Boolean}
		@param requiresrestart: Whether or not the changing of this setting
		requires a restart
		@type requiresrestart: C{Boolean}
		"""
		entry = SettingEntry(module, name, applyfunction, initialdata, requiresrestart)
		self.addEntry(entry)

	def addEntry(self, entry):
		"""Adds a new C{SettingEntry} to the Settting
		@param entry: A new SettingEntry that is to be added
		@type entry: C{SettingEntry}
		"""
		if entry.module not in self._entries:
			self._entries[entry.module] = {}
		self._entries[entry.module][entry.name] = entry

		"""
		# Make sure the new entry is available
		if self.get(entry.module, entry.name) is None:
			print "Updating", self._settings_file, "to the default, it is missing the entry:"\
			      , entry.name ,"for module", entry.module

			#self.setDefaults()
		if self.get(entry.module, entry.name) is None:
			print "WARNING:", entry.module, ":", entry.name, "still not found!"
		"""

	def saveSettings(self, filename=""):
		""" Writes the settings to the settings file

		@param filename: Specifies the file to save the settings to.  If it is not specified
		the original settings file is used.
		@type filename: C{string}
		"""
		if self._serializer:
			if filename == "":
				self._serializer.save(os.path.join(self._appdata, self._settings_file))
			else:
				self._serializer.save(filename)

	# get all the settings of a module name module
	def getSettingsFromFile(self, module, logger=None):
		if self._serializer:

			self._logger = logger
			modules = self._serializer.getModuleNameList()
			self._settingsFromFile[module] = self._serializer.getAllSettings(module)

			if self._logger:
				self._logger.log_log("Loading Settings From File ...")

			if self._settingsFromFile[module] is not None:
				self._readSettingsCompleted[module] = True

			# we need validation for the module FIFE only
			if module is not "FIFE":
				return self._settingsFromFile[module]
			"""
			Now we have all the settings we needed. We have to validate the settings. Applicable for module
			FIFE only
			"""
			for name in self._settingsFromFile[module]:
				# if the setting name is known, so that it is
				# both in self._settingsFromFile and validSetting
				if name in self._validSetting[module]:

					e_value = self._settingsFromFile[module][name]

					if name == "InitialVolume":
						if e_value >= self._validSetting[module][name][0] and e_value <= self._validSetting[module][name][1]:
							self._settingsFromFile[module][name] = e_value
						else:
							if self._logger:
								self._logger.log_log("InitalVolume must have a value between 0.0 and 10.0")

					elif name == "GLAlphaTestValue":
						if e_value >= self._validSetting[module][name][0] and e_value <= self._validSetting[module][name][1]:
							self._settingsFromFile[module][name] = e_value
						else:
							if self._logger:
								self._logger.log_log("GLAlphaTestValue must have a value between 0.0 and 1.0")

					elif name == "ColorKey":
						e_value = e_value.split(',')
						if int(e_value[0]) in range(0,256) and int(e_value[1]) in range(0,256) and int(e_value[2]) in range(0,256):
							self._settingsFromFile[name] = [int(e_value[0]),int(e_value[1]),int(e_value[2])];

						else:
							if self._logger:
								self._logger.log_log("ColorKey values must be within 0 and 255. Setting to Default Value.")

					elif name == "ScreenResolution":
						temp = e_value.split('x')
						if len(temp) == 2:
							self._settingsFromFile[module][name]=e_value
						else:
							if self._logger:
								self._logger.log_log("Invalid Screen Resolution value. We expect two integer separated by x")

					elif len(self._validSetting[module][name]) == 0:
						self._settingsFromFile[module][name] = e_value

					elif name == "LogModules":
						for checking_element in e_value:
							module_valid = False
							for base_element in self._validSetting[module][name]:
								# checking_element is valid

								if checking_element == base_element:
									module_valid = True
									already_in = False
									for element in self._settingsFromFile[module][name]:
										if element == checking_element:
											already_in = True
									if already_in == False:
										self._settingsFromFile[module][name].append(checking_element)
							if module_valid == False:
								if self._logger:
									self._logger.log_log(checking_element +" is not a valid logModule")
					elif name == "FrameLimit":
						if e_value > 0:
							self._settingsFromFile[module][name] = e_value
						else:
							if self._logger:
								self._logger.log_log(e_value + " is not a valid FrameLimit setting.  You must specify a positive integer!")
					elif name == "MouseSensitivity":
						self._settingsFromFile[module][name] = e_value
					elif name == "MouseAcceleration":
						self._settingsFromFile[module][name] = e_value
					elif name == "NativeImageCursor":
						self._settingsFromFile[module][name] = e_value
					elif name == "JoystickSupport":
						self._settingsFromFile[module][name] = e_value

					elif name in ("SDLRemoveFakeAlpha", "LogToPrompt", "LogToFile"):
						if type(e_value) == int:
							try:
								e_value = (False, True)[e_value]
							except IndexError:
								self._logger.log_warn("Invalid int-value for %s. Defaulted to False!"%name)
								e_value = False
							self._logger.log_warn("Use of type int for %s is deprecated. Use bool instead!"%name)
						self._settingsFromFile[module][name] = e_value

					else:

						if isinstance(self._settingsFromFile[module][name],list) == True or isinstance(self._settingsFromFile[module][name],dict) == True:
							valid = False
							for value in self._validSetting[module][name]:
								if value == e_value:
									valid = True
									self._settingsFromFile[module][name] = e_value;
							if valid == False:
								if self._logger:
									self._logger.log_log("Setting " + name + " got invalid value. Setting to Default.")
						else: self._settingsFromFile[module][name] = e_value

				# name is unknown
				else:
					if self._logger:
						self._logger.log_log("Setting "+ name + " is unknown")

			if self._logger:
				self._logger.log_log("Settings Loaded ...")

			"""
			Upto this point we have validated all the settings that are in settings.xml file. But, what if a setting is valid and still it is
			not present in the settings.xml file. For this, we should give them the default Values that are in defaultSetting.
			"""

			for name in self._defaultSetting[module]:
				if name not in self._settingsFromFile[module]:
					self._settingsFromFile[module][name] = self._defaultSetting[module][name]

			return self._settingsFromFile[module]

		else:
			return None

	def get(self, module, name, defaultValue=None):
		""" Gets the value of a specified setting

		@param module: Name of the module to get the setting from
		@param name: Setting name
		@param defaultValue: Specifies the default value to return if the setting is not found
		@type defaultValue: C{str} or C{unicode} or C{int} or C{float} or C{bool} or C{list} or C{dict}
		"""

		if self._serializer:
			if module is "FIFE":
				# check whether getAllSettings has been called already
				if self._readSettingsCompleted[module] is not True:
					value = self._serializer.get(module, name, defaultValue)

					if value is not None:
						return value
					else:
						if name in self._defaultSetting[module]:
							return self._defaultSetting[module][name]
						else:
							raise Exception(str(name) + ' is neither in settings.xml nor it has a default value set')
				else:
					if name in self._settingsFromFile[module]:
						return self._settingsFromFile[module][name]
					else:
						raise Exception(str(name) + ' is neither in settings.xml nor it has a default value set')
			else:
				return self._serializer.get(module, name, defaultValue)
		else:
			"""
			serializer not set, reading from default value
			"""
			if name in self._defaultSetting:
				return self._defaultSetting[module][name]
			else:
				raise Exception(str(name) + ' is neither in settings.xml nor it has a default value set')

	def set(self, module, name, value, extra_attrs={}):
		"""
		Sets a setting to specified value.

		@param module: Module where the setting should be set
		@param name: Name of setting
		@param value: Value to assign to setting
		@type value: C{str} or C{unicode} or C{int} or C{float} or C{bool} or C{list} or C{dict}
		@param extra_attrs: Extra attributes to be stored in the XML-file
		@type extra_attrs: C{dict}
		"""

		#update the setting cache
		if module in self._settingsFromFile:
			self._settingsFromFile[module][name] = value
		else:
			self._settingsFromFile[module] = { name: value }

		if self._serializer:
			self._serializer.set(module, name, value, extra_attrs)

	def remove(self, module, name):
		"""
		Removes a variable

		@param module: Module where the variable should be set
		@param name: Name of the variable
		"""
		#update the setting cache
		if module in self._settingsFromFile:
			del self._settingsFromFile[module][name]

		if self._serializer:
			self._serializer.remove(module, name)

	def setAvailableScreenResolutions(self, reslist):
		"""
		A list of valid default screen resolutions.   This should be called once
		right after you instantiate Settings.

		Valid screen resolutions must be strings in the form of: WIDTHxHEIGHT

		Example:
			settings.setAvailableScreenResolutions(["800x600", "1024x768"])
		"""
		self._resolutions = reslist

	def setDefaults(self):
		"""
		Overwrites the setting file with the default settings file.
		"""
		shutil.copyfile(self._default_settings_file, os.path.join(self._appdata, self._settings_file))
		self.changesRequireRestart = True
		self.initSerializer()

	def _getEntries(self):
		return self._entries

	def _setEntries(self, entries):
		self._entries = entries

	def _getSerializer(self):
		return self._serializer

	entries = property(_getEntries, _setEntries)
	serializer = property(_getSerializer)

class SettingEntry(object):

	def __init__(self, module, name, applyfunction=None, initialdata=None, requiresrestart=False):
		"""
		@param module: The Setting module this Entry belongs to
		@type module: C{String}
		@param name: The Setting's name
		@type name: C{String}
		@param applyfunction: function that makes the changes when the Setting is
		saved
		@type applyfunction: C{function}
		@param initialdata: If the widget supports the setInitialData() function
		this can be used to set the initial data
		@type initialdata: C{String} or C{Boolean}
		@param requiresrestart: Whether or not the changing of this setting
		requires a restart
		@type requiresrestart: C{Boolean}
		"""
		self._module = module
		self._name = name
		self._requiresrestart = requiresrestart
		self._initialdata = initialdata
		self._applyfunction = applyfunction

	def onApply(self, data):
		"""Implement actions that need to be taken when the setting is changed
		here.
		"""
		if self._applyfunction is not None:
			self._applyfunction(data)

	def _getModule(self):
		return self._module

	def _setModule(self, module):
		self._module = module

	def _getName(self):
		return self._name

	def _setName(self, name):
		self._name = name

	def _getRequiresRestart(self):
		return self._requiresrestart

	def _setRequiresRestart(self, requiresrestart):
		self._requiresrestart = requiresrestart

	def _getInitialData(self):
		return self._initialdata

	def _setInitialData(self, initialdata):
		self._initialdata = initialdata

	def _getApplyFunction(self):
		return self._applyfunction

	def _setApplyFunction(self, applyfunction):
		self._applyfunction = applyfunction

	module = property(_getModule, _setModule)
	name = property(_getName, _setName)
	requiresrestart = property(_getRequiresRestart, _setRequiresRestart)
	initialdata = property(_getInitialData, _setInitialData)
	applyfunction = property(_getApplyFunction, _setApplyFunction)

	def __str__(self):
		return "SettingEntry: " +  self.name + " Module: " + self.module +  \
		        " requiresrestart: " + str(self.requiresrestart) + \
		        " initialdata: " + str(self.initialdata)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2013 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from builtins import range
from builtins import object
import weakref
from fife import fifechan
from fife.extensions.pychan.attrs import BoolAttr, IntAttr, PointAttr, MixedListAttr
from .widget import Widget

class PieSegment(object):
	def __init__(self, pie, start, stop, color):
		self.pie = weakref.ref(pie)
		self.start = start
		self.stop = stop
		self._color = None
		self.color = color

	def _setStartAngle(self, angle):
		self.start = angle
		if self.pie is not None:
			self.pie().update()
	def _getStartAngle(self): return self.start
	start_angle = property(_getStartAngle, _setStartAngle)

	def _setStopAngle(self, angle):
		self.stop = angle
		if self.pie is not None:
			self.pie().update()
	def _getStopAngle(self): return self.stop
	stop_angle = property(_getStopAngle, _setStopAngle)

	def _setColor(self, color):
		if isinstance(color, tuple):
			self._color = fifechan.Color(*color)
		else:
			# Force a copy to get value semantics
			self._color = fifechan.Color(color.r,color.g,color.b,color.a)
		if self.pie is not None:
			self.pie().update()
	def _getColor(self):
		return fifechan.Color(self._color.r, self._color.g, self._color.b, self._color.a)
	color = property(_getColor, _setColor)


class PieGraph(Widget):
	""" A pie graph widget

	New Attributes
	==============

	  - coordinates: int list: x and y coordinates
	  - thickness': int: Line thickness, default 1
	  - pie_color: color: Pie color
	  - opaque: bool: default False
	"""

	ATTRIBUTES = Widget.ATTRIBUTES + [ PointAttr('center'),
									   IntAttr('radius'),
									   BoolAttr('opaque'),
									   MixedListAttr('segments')
									 ]
	DEFAULT_HEXPAND = False
	DEFAULT_VEXPAND = False

	DEFAULT_RADIUS = 10
	DEFAULT_OPAQUE = False

	def __init__(self,
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 opaque = None,
				 center = None,
				 radius = None,
				 segments = None):

		self.real_widget = fifechan.PieGraph()
		self.opaque = self.DEFAULT_OPAQUE
		self.radius = self.DEFAULT_RADIUS
		self._segments = []

		super(PieGraph, self).__init__(parent=parent,
									   name=name,
									   size=size,
									   min_size=min_size,
									   max_size=max_size,
									   fixed_size=fixed_size,
									   margins=margins,
									   padding=padding,
									   helptext=helptext,
									   position=position,
									   style=style,
									   hexpand=hexpand,
									   vexpand=vexpand,
									   font=font,
									   base_color=base_color,
									   background_color=background_color,
									   foreground_color=foreground_color,
									   selection_color=selection_color,
									   border_color=border_color,
									   outline_color=outline_color,
									   border_size=border_size,
									   outline_size=outline_size,
									   position_technique=position_technique,
									   is_focusable=is_focusable,
									   comment=comment)

		if opaque is not None: self.opaque = opaque
		if center is not None: self.center = center
		if radius is not None: self.radius = radius
		if segments is not None: self.segments = segments


	def clone(self, prefix):
		pieGraphClone = PieGraph(None,
					self._createNameWithPrefix(prefix),
					self.size,
					self.min_size,
					self.max_size,
					self.fixed_size,
					self.margins,
					self.padding,
					self.helptext,
					self.position,
					self.style,
					self.hexpand,
					self.vexpand,
					self.font,
					self.base_color,
					self.background_color,
					self.foreground_color,
					self.selection_color,
					self.border_color,
					self.outline_color,
					self.border_size,
					self.outline_size,
					self.position_technique,
					self.is_focusable,
					self.comment,
					self.opaque,
					self.center,
					self.radius,
					self.segments)
		return pieGraphClone

	def update(self):
		self.real_widget.clearSegments()
		for s in self._segments:
			self.real_widget.addSegment(s.start_angle, s.stop_angle, s.color)
		
	def _setOpaque(self, opaque): self.real_widget.setOpaque(opaque)
	def _getOpaque(self): return self.real_widget.isOpaque()
	opaque = property(_getOpaque, _setOpaque)

	def _setCenter(self, center):
		# point check
		if isinstance(center, fifechan.Point):
			self.real_widget.setCenter(center)
			return
		self.real_widget.setCenter(center[0], center[1])
	def _getCenter(self): return (self.real_widget.getCenterX(), self.real_widget.getCenterY())
	center = property(_getCenter, _setCenter)

	def _setRadius(self, radius): self.real_widget.setRadius(radius)
	def _getRadius(self): return self.real_widget.getRadius()
	radius = property(_getRadius, _setRadius)

	def addSegment(self, start, stop, color):
		segment = PieSegment(self, start, stop, color)
		self._segments.append(segment)
		self.real_widget.addSegment(segment.start_angle, segment.stop_angle, segment.color)
		
	def removeSegment(self, start, stop, color):
		tmp_color = color
		if isinstance(color, tuple):
			tmp_color = fifechan.Color(*color)

		for s in self._segments:
			if s.start_angle == start and s.stop_angle == stop and s.color == tmp_color:
				self._segments.remove(s)
				self.update()
				break
	
	def clearSegments(self):
		self._segments = []
		self.real_widget.clearSegments()

	def _setSegments(self, segments):
		self._segments = []
		if segments is None:
			self.real_widget.clearSegments()
			return
		if isinstance(segments[0], PieSegment):
			self._segments = segments
		else:
			for i in range(0, len(segments), 3):
				segment = PieSegment(self, segments[i], segments[i+1], segments[i+2])
				self._segments.append(segment)
		self.update()
	def _getSegments(self): return self._segments
	segments = property(_getSegments, _setSegments)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

from __future__ import absolute_import
from fife import fifechan

from fife.extensions.pychan.properties import ImageProperty
from fife.extensions.pychan.attrs import Attr, BoolAttr, IntAttr

from .widget import Widget


class IconProgressBar(Widget):
	"""
	An image icon.

	New Attributes
	==============

	  - image: String or GuiImage: The source location of the Image or a direct GuiImage
	  - max_icons: Maximal count of icons
	  - icons: Current count of active icons
	  - orientation: Horizontal (0) or Vertical (1)
	  - opaque: True if the widget is opaque, false otherwise.
	"""
	ATTRIBUTES = Widget.ATTRIBUTES + [ Attr('image'), 
									   IntAttr('max_icons'),
									   IntAttr('icons'),
									   IntAttr('orientation'),
									   BoolAttr('opaque')
									 ]

	DEFAULT_OPAQUE = True
									 
	def __init__(self, 
				 parent = None,
				 name = None,
				 size = None,
				 min_size = None,
				 max_size = None,
				 fixed_size = None,
				 margins = None,
				 padding = None,
				 helptext = None,
				 position = None,
				 style = None,
				 hexpand = None,
				 vexpand = None,
				 font = None,
				 base_color = None,
				 background_color = None,
				 foreground_color = None,
				 selection_color = None,
				 border_color = None,
				 outline_color = None,
				 border_size = None,
				 outline_size = None,
				 position_technique = None,
				 is_focusable = None,
				 comment = None,
				 image = None,
				 max_icons = None,
				 icons = None,
				 orientation = None,
				 opaque = None):
				 
		self.real_widget = fifechan.IconProgressBar() 
		self.opaque = opaque or self.DEFAULT_OPAQUE
		
		super(IconProgressBar,self).__init__(parent=parent,
											 name=name,
											 size=size,
											 min_size=min_size,
											 max_size=max_size,
											 fixed_size=fixed_size,
											 margins=margins,
											 padding=padding,
											 helptext=helptext,
											 position=position,
											 style=style,
											 hexpand=hexpand,
											 vexpand=vexpand,
											 font=font,
											 base_color=base_color,
											 background_color=background_color,
											 foreground_color=foreground_color,
											 selection_color=selection_color,
											 border_color=border_color,
											 outline_color=outline_color,
											 border_size=border_size,
											 outline_size=outline_size,
											 position_technique=position_technique,
											 is_focusable=is_focusable,
											 comment=comment)
		
		# for the case that image can not be found, e.g. invalid path
		# the IconProgressbar is removed from the manager
		try:
			self.image = image
		except Exception:
			get_manager().removeWidget(self)
			raise

		if max_icons is not None: self.max_icons = max_icons
		if icons is not None: self.icons = icons

	def clone(self, prefix):
		iconClone = IconProgressBar(None,
				self._createNameWithPrefix(prefix),
				self.size,
				self.min_size,
				self.max_size,
				self.fixed_size,
				self.margins,
				self.padding,
				self.helptext,
				self.position,
				self.style,
				self.hexpand,
				self.vexpand,
				self.font,
				self.base_color,
				self.background_color,
				self.foreground_color,
				self.selection_color,
				self.border_color,
				self.outline_color,
				self.border_size,
				self.outline_size,
				self.position_technique,
				self.is_focusable,
				self.comment,
				self.image,
				self.max_icons,
				self.icons,
				self.orientation,
				self.opaque)
				 
		
		return iconClone
		
	_image = ImageProperty("Image")

	def advance(self):
		self.real_widget.advance()
		
	def reset(self):
		self.real_widget.reset()
	
	def _setImage(self,source):
		self._image = source

	def _getImage(self):
		return self._image
	image = property(_getImage,_setImage)
	
	def _setMaxIcons(self, maxIcons):
		if type(maxIcons) != int:
			raise RuntimeError("IconProgressBar max icons should be an integer")
		self.real_widget.setMaxIcons(maxIcons)
		
	def _getMaxIcons(self):
		return self.real_widget.getMaxIcons()
	max_icons = property(_getMaxIcons, _setMaxIcons)

	def _setIcons(self, icons):
		if type(icons) != int:
			raise RuntimeError("IconProgressBar icons should be an integer")
		self.real_widget.setIconCount(icons)	
	def _getIcons(self):
		return self.real_widget.getIconCount()
	icons = property(_getIcons, _setIcons)
	
	def _setOrientation(self, orientation):
		self.real_widget.setOrientation(orientation)
	
	def _getOrientation(self):
		return self.real_widget.getOrientation()
	orientation = property(_getOrientation, _setOrientation)
	
	def _setOpaque(self,opaque): 
		self.real_widget.setOpaque(opaque)
	def _getOpaque(self): 
		return self.real_widget.isOpaque()
	opaque = property(_getOpaque,_setOpaque)



# -*- coding: utf-8 -*-

# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

"""
Functional utilities designed for pychan use cases.
"""
from __future__ import print_function
from __future__ import absolute_import

import sys

from builtins import range
from . import exceptions

### Functools ###

def applyOnlySuitable(func,*args,**kwargs):
	"""
	This nifty little function takes another function and applies it to a dictionary of
	keyword arguments. If the supplied function does not expect one or more of the
	keyword arguments, these are silently discarded. The result of the application is returned.
	This is useful to pass information to callbacks without enforcing a particular signature.
	"""
	if sys.version_info < (3,):
		func_name = 'im_func'
		code_name = 'func_code'
	else:
		func_name = '__func__'
		code_name = '__code__'
	if hasattr(func, func_name):
		code = func.__func__.__code__
		varnames = code.co_varnames[1:code.co_argcount]#ditch bound instance
	elif hasattr(func, code_name):
		code = func.__code__
		varnames = code.co_varnames[0:code.co_argcount]
	elif hasattr(func,'__call__'):
		func = func.__call__
		if hasattr(func, func_name):
			code = func.__func__.__code__
			varnames = code.co_varnames[1:code.co_argcount]#ditch bound instance
		elif hasattr(func, code_name):
			code = func.__code__
			varnames = code.co_varnames[0:code.co_argcount]

	#http://docs.python.org/lib/inspect-types.html
	if code.co_flags & 8:
		return func(*args,**kwargs)
	for name,value in list(kwargs.items()):
		if name not in varnames:
			del kwargs[name]
	return func(*args,**kwargs)

def callbackWithArguments(callback,*args,**kwargs):
	"""
	Curries a function with extra arguments to
	create a suitable callback.

	If you don't know what this means, don't worry.
	It is designed for the case where you need
	different buttons to execute basically the same code
	with different arguments.

	Usage::
	  # The target callback
	  def printStuff(text):
	      print text
	  # Mapping the events
	  gui.mapEvents({
	      'buttonHello' : callbackWithArguments(printStuff,"Hello"),
	      'buttonBye' : callbackWithArguments(printStuff,"Adieu")
	  })
	"""
	def real_callback():
		callback(*args,**kwargs)
	return real_callback

def attrSetCallback(**kwargs):
	"""
	Generates an event callback that sets attributes on the widget
	it is called on. This is especially useful for mouseEntered/Exited
	events - to create hover effects.

	It takes a set of keyword arguments. The keys are treated as attribute names,
	which are then set to the corresponding value when the callback is called.
	Some key names are treated special - see below.

	Usage - Example adapted from demo application::
		eventMap = {
			'creditsLink'  : showCreditsCallback,
			'creditsLink/mouseEntered' : attrSetCallback(
			      text = "Show credits!",
			      background_color = (255,255,0,255),
			      do__adaptLayout=True),
			'creditsLink/mouseExited'  : attrSetCallback(text = "Credits"),
		gui.mapEvents(eventMap)

	Now when the mouse enters the creditsLink (a Label in our case), the following code will be executed::
		#widget is the creditsLink - given to the event callback.
		widget.text = "Show credits!"
		widget.background_color = (255,255,0,255)
		widget.adaptLayout()

	The C{do__adaptLayout} argument causes the method C{adaptLayout} to be called.
	In fact any key starting with C{do__} results in such a method call. The order
	of execution of such calls is undefined.

	Keys starting with an underscore raise a L{exceptions.PrivateFunctionalityError}.
	"""
	do_calls = []

	for name in list(kwargs.keys()):
		if name.startswith("_"):
			raise exceptions.PrivateFunctionalityError(name)
		if name.startswith("do__"):
			do_calls.append(name[4:])
			del kwargs[name]

	def attrSet_callback(widget=None):
		for name,value in list(kwargs.items()):
			setattr(widget,name,value)
		for method_name in do_calls:
			method = getattr(widget,method_name)
			method()
	return attrSet_callback

def chainCallbacks(*args):
	"""
	Chains callbacks to be called one after the other.
	
	Example Usage::
	    def print_event(event=0):
	      print event
	    def print_widget(widget=0):
	      print widget
            callback = tools.chainCallbacks(doSomethingUseful, print_event, print_widget)
	    guiElement.capture(callback)
	"""
	callbacks = args
	def chain_callback(event=0,widget=0):
		for callback in callbacks:
			applyOnlySuitable(callback, event=event, widget=widget)
	return chain_callback

def repeatALot(n = 1000):
	"""
	Internal decorator used to profile some pychan functions.
	Only use with functions without side-effect.

	Usage::
		@repeatALot(n=10000)
		def findChild(self,**kwargs):
			...
	"""
	def wrap_f(f):
		def new_f(*args,**kwargs):
			for i in range(n):
				f(*args,**kwargs)
			return f(*args,**kwargs)
		return new_f
	return wrap_f

def this_is_deprecated(func,message=None):
	if message is None:
		message = repr(func)
	def wrapped_func(*args,**kwargs):
		print("PyChan: You are using the DEPRECATED functionality: %s" % message)
		return func(*args,**kwargs)
	return wrapped_func



# -*- coding: utf-8 -*-
# ####################################################################
#  Copyright (C) 2005-2019 by the FIFE team
#  http://www.fifengine.net
#  This file is part of FIFE.
#
#  FIFE is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the
#  Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
# ####################################################################

""" main xml parser class for xml map loading """
from __future__ import print_function

from builtins import str
from builtins import object
import time

from fife import fife

from fife.extensions.serializers import ET
from fife.extensions.serializers import SerializerError, InvalidFormat
from fife.extensions.serializers import NameClash, NotFound, WrongFileType

from fife.extensions.serializers.xmlobject import XMLObjectLoader
from fife.extensions.serializers.xmlanimation import loadXMLAnimation
from fife.extensions.serializers.xml_loader_tools import loadImportFile, loadImportDir
from fife.extensions.serializers.xml_loader_tools import loadImportDirRec
from fife.extensions.serializers.xml_loader_tools import root_subfile, reverse_root_subfile


FORMAT = '1.0'

class XMLMapLoader(object):
	""" The B{XMLMapLoader} parses the xml map using several section.
	Each section fires a callback (if given) which can e. g. be
	used to show a progress bar.

	The callback sends two values, a string and a float (which shows
	the overall process): callback(string, float)
	"""
	def __init__(self, engine, callback, debug, extensions):
		"""
		@type	engine:		object
		@param	engine:		a pointer to fife.engine
		@type	callback:	function
		@param	callback:	a callback with two arguments, optional
		@type	debug:		bool
		@param	debug:		flag to activate / deactivate print statements
		@type	extensions:	dict
		@param	extensions:	information package which extension should be activated (lights, sounds)
		"""
#		self.thisown = 0

		self.callback = callback
		self.debug = debug

		self.engine = engine
		self.vfs = self.engine.getVFS()
		self.model = self.engine.getModel()
		self.image_manager = self.engine.getImageManager()
		self.anim_pool = None

		self.obj_loader = XMLObjectLoader(engine)

		self.map = None
		self.source = None
		self.time_to_load = 0

		self.nspace = None

		self.msg = {}
		self.msg['map'] = 'created map'
		self.msg['imports'] = 'loaded imports'
		self.msg['layer'] = 'loaded layer: %s'
		self.msg['camera'] = 'loaded camera: %s'

		if 'sound' not in extensions:
			extensions['sound'] = False
		if 'lights' not in extensions:
			extensions['lights'] = False

		self.light_data = {}
		self.extensions = extensions

	def _err(self, msg):
		raise SyntaxError(''.join(['File: ', self.source, ' . ', msg]))

	def loadResource(self, location):
		""" overwrite of B{fife.ResourceLoader}

		@type	location:	object
		@param	location:	path to a map file as a fife.ResourceLocation
		@return	FIFE map object
		@rtype	object
		"""
		start_time = time.time()
		self.source = location
		f = self.vfs.open(self.source)
		f.thisown = 1
		tree = ET.parse(f)
		root = tree.getroot()

		map = self.parse_map(root)
		self.time_to_load = time.time() - start_time
		return map

	def parse_map(self, mapelt):
		""" start parsing the xml structure and
		call submethods for turning found tags
		into FIFE objects and create the map

		@type	mapelt:	object
		@param	mapelt:	ElementTree root
		@return	FIFE map object
		@rtype	object
		"""
		if not mapelt:
			self._err('No <map> element found at top level of map file definition.')
		_id, format = mapelt.get('id'), mapelt.get('format')

		if not format == FORMAT: self._err(''.join(['This file has format ', format, ' but this loader has format ', FORMAT]))
		if not _id: self._err('Map declared without an identifier.')

		map = None
		try:
			self.map = self.model.createMap(str(_id))
			self.map.setFilename(self.source)
		except fife.Exception as e: # NameClash appears as general fife.Exception; any ideas?
			print(e.getMessage())
			print(''.join(['File: ', self.source, '. The map ', str(_id), ' already exists! Ignoring map definition.']))
			return map

		# xml-specific directory imports. This is used by xml savers.
		self.map.importDirs = []

		if self.callback is not None:
			self.callback(self.msg['map'], float(0.25) )

		self.parse_imports(mapelt, self.map)
		self.parse_layers(mapelt, self.map)
		self.parse_cameras(mapelt, self.map)

		# create light nodes
		if self.light_data:
			self.create_light_nodes(self.map)

		return self.map

	def parse_imports(self, mapelt, map):
		""" load all objects defined as import into memory

		@type	mapelt:	object
		@param	mapelt:	ElementTree root
		@return	FIFE map object
		@rtype	object
		"""
		parsedImports = {}

		if self.callback:
			tmplist = mapelt.findall('import')
			i = float(0)

		for item in mapelt.findall('import'):
			_file = item.get('file')
			if _file:
				_file = reverse_root_subfile(self.source, _file)
			_dir = item.get('dir')
			if _dir:
				_dir = reverse_root_subfile(self.source, _dir)

			# Don't parse duplicate imports
			if (_dir,_file) in parsedImports:
				if self.debug: print("Duplicate import:" ,(_dir, _file))
				continue
			parsedImports[(_dir,_file)] = 1

			if _file and _dir:
				loadImportFile(self.obj_loader, '/'.join(_dir, _file), self.engine, self.debug)
			elif _file:
				loadImportFile(self.obj_loader, _file, self.engine, self.debug)
			elif _dir:
				loadImportDirRec(self.obj_loader, _dir, self.engine, self.debug)
				map.importDirs.append(_dir)
			else:
				if self.debug: print('Empty import statement?')

			if self.callback:
				i += 1
				self.callback(self.msg['imports'], float( i / float(len(tmplist)) * 0.25 + 0.25 ) )

	def parse_layers(self, mapelt, map):
		""" create all layers and their instances

		@type	mapelt:	object
		@param	mapelt:	ElementTree root
		@type	map:	object
		@param	map:	FIFE map object
		"""
		if self.callback is not None:
			tmplist = mapelt.findall('layer')
			i = float(0)

		for layer in mapelt.findall('layer'):
			_id = layer.get('id')
			grid_type = layer.get('grid_type')

			if not _id: self._err('<layer> declared with no id attribute.')
			if not grid_type: self._err(''.join(['Layer ', str(_id), ' has no grid_type attribute.']))

			x_scale = layer.get('x_scale')
			y_scale = layer.get('y_scale')
			rotation = layer.get('rotation')
			x_offset = layer.get('x_offset')
			y_offset = layer.get('y_offset')
			z_offset = layer.get('z_offset')
			pathing = layer.get('pathing')
			transparency = layer.get('transparency')

			layer_type = layer.get('layer_type')
			layer_type_id = layer.get('layer_type_id')

			if not x_scale: x_scale = 1.0
			if not y_scale: y_scale = 1.0
			if not rotation: rotation = 0.0
			if not x_offset: x_offset = 0.0
			if not y_offset: y_offset = 0.0
			if not z_offset: z_offset = 0.0
			if not pathing: pathing = "cell_edges_only"
			if not transparency:
				transparency = 0
			else:
				transparency = int(transparency)

			cellgrid = self.model.getCellGrid(grid_type)
			if not cellgrid: self._err('<layer> declared with invalid cellgrid type. (%s)' % grid_type)

			cellgrid.setRotation(float(rotation))
			cellgrid.setXScale(float(x_scale))
			cellgrid.setYScale(float(y_scale))
			cellgrid.setXShift(float(x_offset))
			cellgrid.setYShift(float(y_offset))
			cellgrid.setZShift(float(z_offset))

			layer_obj = None
			try:
				layer_obj = map.createLayer(str(_id), cellgrid)
			except fife.Exception as e:
				print(e.getMessage())
				print('The layer ' + str(_id) + ' already exists! Ignoring this layer.')

				continue

			strgy = fife.CELL_EDGES_ONLY
			if pathing == "cell_edges_and_diagonals":
				strgy = fife.CELL_EDGES_AND_DIAGONALS

			layer_obj.setPathingStrategy(strgy)
			layer_obj.setLayerTransparency(transparency)

			if layer_type:
				if layer_type == 'walkable':
					layer_obj.setWalkable(True)
				elif layer_type == 'interact' and layer_type_id:
					layer_obj.setInteract(True, layer_type_id)

			self.parse_instances(layer, layer_obj)

			if self.extensions['lights']:
				self.parse_lights(layer, layer_obj)
			if self.extensions['sound']:
				self.parse_sounds(layer, layer_obj)

			if self.callback is not None:
				i += 1
				self.callback(self.msg['layer'] % str(_id), float( i / float(len(tmplist)) * 0.25 + 0.5 ) )


		layers = map.getLayers()
		for l in layers:
			if l.isInteract():
				walk_layer = map.getLayer(l.getWalkableId())
				if walk_layer:
					walk_layer.addInteractLayer(l);

		for l in layers:
			if l.isWalkable():
				l.createCellCache()

		# cleanup
		if self.callback is not None:
			del tmplist
			del i

	def parse_lights(self, layerelt, layer):
		""" create light nodes

		@type	layerelt:	object
		@param	layerelt:	ElementTree layer branch
		@type	layer:	object
		@param	layer:	FIFE layer object
		"""
		_LIGHT_DEFAULT_BLENDING_SRC = -1
		_LIGHT_DEFAULT_BLENDING_DST = -1
		_LIGHT_DEFAULT_SUBDIVISIONS = 32
		_LIGHT_DEFAULT_CAM_ID = 'default'
		_LIGHT_DEFAULT_INTENSITY = 128
		_LIGHT_DEFAULT_RADIUS = 10.0

		print("Processing lights ... ")
		lightelt = layerelt.find('lights')
		if not lightelt:
			print("\tno lights found on layer %s" % layer.getId())
			return

		lights = []
		for attr in ('l', 'light', 'lgt'):
			lights.extend(lightelt.findall(attr))

		for light in lights:
			group = light.get('group')
			if not group:
				print("Light has no group. Omitting...")
				continue

			blending_src = light.get('src')
			if not blending_src:
				blending_src = _LIGHT_DEFAULT_BLENDING_SRC
			blending_dst = light.get('dst')
			if not blending_dst:
				blending_dst = _LIGHT_DEFAULT_BLENDING_DST

			_x = light.get('x')
			if not _x: _x = 0
			_y = light.get('y')
			if not _y: _y = 0
			_z = light.get('y')
			if not _z: _z = 0

			node = {}
			node['blending_src'] = int(blending_src)
			node['blending_dst'] = int(blending_dst)
			node['layer'] = layer.getId()
			node['position'] = int(_x), int(_y), int(_z)

			# where is the light? *sing*
			instance_id = light.get('instance')
			node['instance'] = None
			if instance_id and layer.getInstance(instance_id):
				node['instance'] = instance_id

			type = light.get('type')
			if type:
				s_ref = light.get('s_ref')
				if not s_ref: s_ref = -1
				node['s_ref'] = int(s_ref)
				a_ref = light.get('a_ref')
				if not a_ref: a_ref = 0.0
				node['a_ref'] = float(a_ref)

				if type == 'image':
					image = light.get('image')
					if not image:
						print("Light has no image. Omitting...")
						continue
					node['type'] = 'image'
					image = reverse_root_subfile(self.source, image)
					img = self.image_manager.create(image)
					node['image'] = img
				elif type == 'animation':
					animation = light.get('animation')
					if not animation:
						print("Light has no animation. Omitting...")
						continue
					node['type'] = 'animation'
					animation = reverse_root_subfile(self.source, animation)
					anim = loadXMLAnimation(self.engine, animation)
					node['animation'] = anim
				elif type == 'simple':
					node['type'] = type
					radius = light.get('radius')
					if not radius: radius = _LIGHT_DEFAULT_RADIUS
					node['radius'] = float(radius)

					subdivisions = light.get('subdivisions')
					if not subdivisions:
						subdivisions = _LIGHT_DEFAULT_SUBDIVISIONS
					node['subdivisions'] = int(subdivisions)

					intensity = light.get('intensity')
					if not intensity:
						intensity = _LIGHT_DEFAULT_INTENSITY
					node['intensity'] = int(intensity)

					xstretch = light.get('xstretch')
					if not xstretch: xstretch = 1.0
					ystretch = light.get('ystretch')
					if not ystretch: ystretch = 1.0
					node['stretch'] = float(xstretch), float(ystretch)

					color = light.get('color')
					if not color: color = '%d,%d,%d' % (255, 255, 255)
					node['color'] = ([int(c) for c in color.split(',')])

			else:
				continue

			cam_id = light.get('camera_id')
			if not cam_id: cam_id = _LIGHT_DEFAULT_CAM_ID

			if not cam_id in self.light_data:
				self.light_data[cam_id] = {}
			if group not in self.light_data[cam_id]:
				self.light_data[cam_id][group] = []

			self.light_data[cam_id][group].append(node)

		for camera, groups in self.light_data.items():
			print("Lights for camera %s" % camera)
			for group, lights in groups.items():
				print(group, lights)

	def parse_sounds(self, layerelt, layer):
		""" create sound emitter

		FIXME:
			- FIFE has a hard limit of sound emitters
			  how should we load emitters here?
			- my first thought: collect a list of sound
			  files & data for emitter creation,
			  then let the client decide what to do with it

		@type	layerelt:	object
		@param	layerelt:	ElementTree layer branch
		@type	layer:	object
		@param	layer:	FIFE layer object
		"""
		# to be continued
		pass

	def parse_instances(self, layerelt, layer):
		""" create all layers and their instances

		@type	layerelt:	object
		@param	layerelt:	ElementTree layer branch
		@type	layer:	object
		@param	layer:	FIFE layer object
		"""
		instelt = layerelt.find('instances')

		instances = []
		for attr in ('i', 'inst', 'instance'):
			instances.extend(instelt.findall(attr))

		for instance in instances:
			_id = instance.get('id')
			if not _id:
				_id = ''

			objectID = ''
			for attr in ('o', 'object', 'obj'):
				objectID = instance.get(attr)
				if objectID: break
			if not objectID: self._err('<instance> %s does not specify an object attribute.' % str(objectID))
			objectID = str(objectID)

			nspace = ''
			for attr in ('namespace', 'ns'):
				nspace = instance.get(attr)
				if nspace: break
			# try to reuse the previous namespace
			if not nspace and self.nspace:
				nspace = self.nspace
			if not nspace and not self.nspace: self._err('<instance> %s does not specify an object namespace, and no default is available.' % str(objectID))
			nspace = str(nspace)
			self.nspace = nspace

			# check if there is an object for this instance available, if not -> skip this one
			object = self.model.getObject(objectID, nspace)
			if not object:
				print("Object with id=%s, ns=%s could not be found. Omitting..." % (objectID, nspace))
				continue

			x = instance.get('x')
			if x: self.x = x = float(x)
			else: x = self.x

			y = instance.get('y')
			if y: self.y = y = float(y)
			else: y = self.y

			z = instance.get('z')
			if z: z = float(z)
			else: z = 0.0

			inst = layer.createInstance(object, fife.ExactModelCoordinate(x,y,z), _id)

			rotation = 0
			for attr in ('r', 'rotation'):
				rotation = instance.get(attr)
				if rotation: break
			if not rotation:
				angles = object.get2dGfxVisual().getStaticImageAngles()
				if angles:
					rotation = angles[0]
				else:
					rotation = 0
			else:
				rotation = int(rotation)
			inst.setRotation(rotation)

			over_block = instance.get('override_blocking')
			if over_block is not None:
				inst.setOverrideBlocking(bool(over_block))
				blocking = instance.get('blocking')
				if blocking is not None:
					inst.setBlocking(bool(int(blocking)))

			fife.InstanceVisual.create(inst)

			stackpos = instance.get('stackpos')
			if stackpos:
				inst.get2dGfxVisual().setStackPosition(int(stackpos))

			if (object.getAction('default')):
				target = fife.Location(layer)
				inst.actRepeat('default', target)

	def parse_cameras(self, mapelt, map):
		""" create all cameras and activate them

		FIXME:
			- should the cameras really be enabled here?
			  IMO that's part of the setup within a client
			  (we just _load_ things here)

		@type	mapelt:	object
		@param	mapelt:	ElementTree root
		@type	map:	object
		@param	map:	FIFE map object
		"""
		if self.callback:
			tmplist = mapelt.findall('camera')
			i = float(0)

		for camera in mapelt.findall('camera'):
			_id = camera.get('id')
			zoom = camera.get('zoom')
			tilt = camera.get('tilt')
			rotation = camera.get('rotation')
			ref_cell_width = camera.get('ref_cell_width')
			ref_cell_height = camera.get('ref_cell_height')
			viewport = camera.get('viewport')
			light_color = camera.get('light_color')

			if not zoom: zoom = 1
			if not tilt: tilt = 0
			if not rotation: rotation = 0

			if not _id: self._err('Camera declared without an id.')
			if not (ref_cell_width and ref_cell_height): self._err(''.join(['Camera ', str(_id), ' declared without reference cell dimensions.']))

			try:
				if viewport:
					cam = map.addCamera(str(_id), fife.Rect(*[int(c) for c in viewport.split(',')]))

				else:
					screen = self.engine.getRenderBackend()
					cam = map.addCamera(str(_id), fife.Rect(0,0,screen.getScreenWidth(),screen.getScreenHeight()))

				renderer = fife.InstanceRenderer.getInstance(cam)
				renderer.activateAllLayers(map)

			except fife.Exception as e:
				print(e.getMessage())

			if light_color: cam.setLightingColor(*[float(c) for c in light_color.split(',')])
			cam.setCellImageDimensions(int(ref_cell_width), int(ref_cell_height))
			cam.setRotation(float(rotation))
			cam.setTilt(float(tilt))
			cam.setZoom(float(zoom))

			renderer = fife.InstanceRenderer.getInstance(cam)
			renderer.activateAllLayers(map)

			if self.callback:
				i += 1
				self.callback(self.msg['camera'] % str(_id), float( i / len(tmplist) * 0.25 + 0.75 ) )

	def create_light_nodes(self, map):
		""" loop through all preloaded lights and create them
		according to their data

		@type	map:	object
		@param	map:	FIFE map object
		"""
		cameras = [i.getId() for i in map.getCameras()]
		renderers = {}
		default_cam = map.getCameras()[0].getId()

		def add_simple_light(group, renderer, node, data):
			""" add a node as simple light to the renderer

			@type	group:	string
			@param	group:	name of the light group
			@type	renderer:	object
			@param	renderer:	fife.LightRenderer instance
			@type	node:		object
			@param	node:		fife.RendererNode instance
			@type	data:		dict
			@param	data:		all data for the light type creation
			"""
			if not node: return
			if not group: return
			renderer.addSimpleLight(
				group,
				node,
				data['intensity'],
				data['radius'],
				data['subdivisions'],
				data['stretch'][0],
				data['stretch'][1],
				data['color'][0],
				data['color'][1],
				data['color'][2],
				data['blending_src'],
				data['blending_dst'],
			)
			if data['s_ref'] is not -1:
				add_stencil_test(group, renderer, data)
		def add_animated_lightmap(group, renderer, node, data):
			""" add a node as animated lightmap to the renderer

			@type	group:	string
			@param	group:	name of the light group
			@type	renderer:	object
			@param	renderer:	fife.LightRenderer instance
			@type	node:		object
			@param	node:		fife.RendererNode instance
			@type	data:		dict
			@param	data:		all data for the light type creation
			"""
			if not node: return
			if not group: return
			renderer.addAnimation(
				group,
				node,
				data['animation'],
				data['blending_src'],
				data['blending_dst'],
			)
			if data['s_ref'] is not -1:
				add_stencil_test(group, renderer, data)
		def add_lightmap(group, renderer, node, data):
			""" add a node as lightmap to the renderer

			@type	group:	string
			@param	group:	name of the light group
			@type	renderer:	object
			@param	renderer:	fife.LightRenderer instance
			@type	node:		object
			@param	node:		fife.RendererNode instance
			@type	data:		dict
			@param	data:		all data for the light type creation
			"""
			if not node: return
			if not group: return
			renderer.addImage(
				group,
				node,
				data['image'],
				data['blending_src'],
				data['blending_dst'],
			)
			if data['s_ref'] is not -1:
				add_stencil_test(group, renderer, data)
		def add_stencil_test(group, renderer, data):
			""" add a stencil test to a group

			@type	group:	string
			@param	group:	name of the light group
			@type	renderer:	object
			@param	renderer:	fife.LightRenderer instance
			@type	data:		dict
			@param	data:		all data for the light type creation
			"""
			if not group: return
			renderer.addStencilTest(
				group,
				data['s_ref'],
				data['a_ref'],
			)

		def create_node(instance=None, point=None, layer=None):
			""" creates a node of one of these types:

				- attached to an instance
				- attached to an instance with offset
				- attached to a point

				FIXME:
					- add location node

			@type:	instance:	object
			@param	instance:	fife instance object
			@type	point:		tuple
			@param	point:		x,y,z tuple
			@type	layer:		object
			@param	layer:		fife layer object
			"""
			node = None
			if not layer: return node

			# node at layer coordinates
			if point and not instance:
				point = fife.Point(point[0], point[1])
				node = fife.RendererNode(point);
			# node with offset
			if instance and point:
				node = fife.RendererNode(instance, layer, fife.Point(point[0], point[1]))
			# node attached to instance
			if instance and not point:
				node = fife.RendererNode(instance, layer)

			if node:
				node.thisown = 0

			return node

		def dump_data():
			""" dump all loaded data """
			for camera, groups in self.light_data.items():
				print("Lights for camera %s" % camera)
				for group, lights in groups.items():
					print(group, lights)

		# fetch all renderer instances for available cameras
		for _id in cameras:
			camera = map.getCamera(_id)
			renderers[_id] = fife.LightRenderer.getInstance(camera)

		# parse data and create the lights
		for camera, groups in self.light_data.items():
			for group, lights in groups.items():
				for light in lights:
					instance = None
					layer = map.getLayer(light['layer'])
					if light['instance']:
						instance = layer.getInstance(light['instance'])
					position = light['position']
					node = create_node(instance, position, layer)

					# some guards
					if not node: continue

					if camera == 'default':
						renderer = renderers[default_cam]
					else:
						renderer = renderers[camera]

					if light['type'] == 'simple':
						add_simple_light(group, renderer, node, light)
					elif light['type'] == 'image':
						add_lightmap(group, renderer, node, light)
					elif light['type'] == 'animation':
						add_animated_lightmap(group, renderer, node, light)

#		dump_data()


